from typing import Any, Dict

import numpy as np
import pytest
import torch
from mmcv.parallel import DataContainer
from PIL import Image

from otx.algorithms.segmentation.adapters.mmseg.datasets.pipelines.transforms import (
    BranchImage,
    DefaultFormatBundle,
    NDArrayToPILImage,
    Normalize,
    PILImageToNDArray,
    RandomColorJitter,
    RandomGaussianBlur,
    RandomGrayscale,
    RandomResizedCrop,
    RandomSolarization,
    TwoCropTransform,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@pytest.fixture(scope="module")
def inputs_np():
    return {
        "img": np.random.randint(0, 10, (16, 16, 3), dtype=np.uint8),
        "gt_semantic_seg": np.random.rand(16, 16),
        "flip": True,
    }


@pytest.fixture(scope="module")
def inputs_PIL():
    return {
        "img": Image.fromarray(np.random.randint(0, 10, (16, 16, 3), dtype=np.uint8)),
        "gt_semantic_seg": np.random.randint(0, 5, (16, 16), dtype=np.uint8),
        "seg_fields": ["gt_semantic_seg"],
        "ori_shape": (16, 16, 3),
    }


class TestNDArrayToPILImage:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:
        self.results: dict = {"img": np.random.randint(0, 255, (3, 3, 3), dtype=np.uint8)}
        self.nd_array_to_pil_image: NDArrayToPILImage = NDArrayToPILImage(keys=["img"])

    @e2e_pytest_unit
    def test_call(self) -> None:
        converted_img: dict = self.nd_array_to_pil_image(self.results)
        assert "img" in converted_img
        assert isinstance(converted_img["img"], Image.Image)

    @e2e_pytest_unit
    def test_repr(self) -> None:
        assert str(self.nd_array_to_pil_image) == "NDArrayToPILImage"


class TestPILImageToNDArray:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:
        self.results: dict = {"img": Image.new("RGB", (3, 3))}
        self.pil_image_to_nd_array: PILImageToNDArray = PILImageToNDArray(keys=["img"])

    @e2e_pytest_unit
    def test_call(self) -> None:
        converted_array: dict = self.pil_image_to_nd_array(self.results)
        assert "img" in converted_array
        assert isinstance(converted_array["img"], np.ndarray)

    @e2e_pytest_unit
    def test_repr(self) -> None:
        assert str(self.pil_image_to_nd_array) == "PILImageToNDArray"


class TestRandomResizedCrop:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:
        self.results: dict = {"img": Image.new("RGB", (10, 16)), "img_shape": (10, 16), "ori_shape": (10, 16)}
        self.random_resized_crop: RandomResizedCrop = RandomResizedCrop((5, 5), (0.5, 1.0))

    @e2e_pytest_unit
    def test_call(self) -> None:
        cropped_img: dict = self.random_resized_crop(self.results)
        assert cropped_img["img_shape"] == (5, 5)
        assert cropped_img["ori_shape"] == (10, 16)


class TestRandomSolarization:
    @pytest.fixture(autouse=True)
    def setUp(self) -> None:
        self.results: dict = {"img": np.random.randint(0, 255, (3, 3, 3), dtype=np.uint8)}
        self.random_solarization: RandomSolarization = RandomSolarization(p=1.0)

    @e2e_pytest_unit
    def test_call(self) -> None:
        solarized: dict = self.random_solarization(self.results)
        assert "img" in solarized
        assert isinstance(solarized["img"], np.ndarray)

    @e2e_pytest_unit
    def test_repr(self) -> None:
        assert str(self.random_solarization) == "RandomSolarization"


class TestNormalize:
    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "mean,std,to_rgb,expected",
        [
            (1.0, 1.0, True, np.array([[[1.0, 0.0, 0.0]]], dtype=np.float32)),
            (1.0, 1.0, False, np.array([[[-1.0, 0.0, 0.0]]], dtype=np.float32)),
        ],
    )
    def test_call(self, mean: float, std: float, to_rgb: bool, expected: np.array) -> None:
        """Test __call__."""
        normalize = Normalize(mean=mean, std=std, to_rgb=to_rgb)
        inputs = dict(img=np.arange(3).reshape(1, 1, 3))

        results = normalize(inputs.copy())

        assert "img" in results
        assert "img_norm_cfg" in results
        assert np.all(results["img"] == expected)

    @e2e_pytest_unit
    @pytest.mark.parametrize("mean,std,to_rgb", [(1.0, 1.0, True)])
    def test_repr(self, mean: float, std: float, to_rgb: bool) -> None:
        """Test __repr__."""
        normalize = Normalize(mean=mean, std=std, to_rgb=to_rgb)

        assert repr(normalize) == normalize.__class__.__name__ + f"(mean={mean}, std={std}, to_rgb=" f"{to_rgb})"


class TestDefaultFormatBundle:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.default_format_bundle = DefaultFormatBundle()

    @e2e_pytest_unit
    @pytest.mark.parametrize("img", [np.ones((1, 1)), np.ones((1, 1, 1)), np.ones((1, 1, 1, 1))])
    @pytest.mark.parametrize("gt_semantic_seg,pixel_weights", [(np.ones((1, 1)), np.ones((1, 1)))])
    def test_call(self, img: np.array, gt_semantic_seg: np.array, pixel_weights: np.array) -> None:
        """Test __call__."""
        inputs = dict(img=img, gt_semantic_seg=gt_semantic_seg, pixel_weights=pixel_weights)

        results = self.default_format_bundle(inputs.copy())

        assert isinstance(results, dict)
        assert "img" in results
        assert isinstance(results["img"], DataContainer)
        assert len(results["img"].data.shape) >= 3
        assert results["img"].data.dtype == torch.float32
        assert "gt_semantic_seg" in results
        assert len(results["gt_semantic_seg"].data.shape) == len(inputs["gt_semantic_seg"].shape) + 1
        assert results["gt_semantic_seg"].data.dtype == torch.int64
        assert "pixel_weights" in results
        assert len(results["pixel_weights"].data.shape) == len(inputs["pixel_weights"].shape) + 1
        assert results["pixel_weights"].data.dtype == torch.float32

    @e2e_pytest_unit
    @pytest.mark.parametrize("img", [np.ones((1,))])
    def test_call_invalid_shape(self, img: np.array):
        inputs = dict(img=img)

        with pytest.raises(ValueError):
            self.default_format_bundle(inputs.copy())

    @e2e_pytest_unit
    def test_repr(self) -> None:
        """Test __repr__."""
        assert repr(self.default_format_bundle) == self.default_format_bundle.__class__.__name__


class TestBranchImage:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.branch_image = BranchImage(key_map={"key1": "key2"})

    @e2e_pytest_unit
    def test_call(self) -> None:
        """Test __call__."""
        inputs = dict(key1="key1", img_fields=["key1"])

        results = self.branch_image(inputs.copy())

        assert isinstance(results, dict)
        assert "key2" in results
        assert results["key1"] == results["key2"]
        assert "key2" in results["img_fields"]

    @e2e_pytest_unit
    def test_repr(self) -> None:
        """Test __repr__."""
        assert repr(self.branch_image) == self.branch_image.__class__.__name__


class TestTwoCropTransform:
    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        mocker.patch(
            "otx.algorithms.segmentation.adapters.mmseg.datasets.pipelines.transforms.build_from_cfg",
            return_value=lambda x: x,
        )
        self.two_crop_transform = TwoCropTransform(view0=[], view1=[])

    @e2e_pytest_unit
    def test_call(self, mocker, inputs_np: Dict[str, Any]) -> None:
        """Test __call__."""
        results = self.two_crop_transform(inputs_np)

        assert isinstance(results, dict)
        assert "img" in results and results["img"].ndim == 4
        assert "gt_semantic_seg" in results and results["gt_semantic_seg"].ndim == 3
        assert "flip" in results and isinstance(results["flip"], list)

    @e2e_pytest_unit
    def test_call_with_single_pipeline(self, mocker, inputs_np: Dict[str, Any]) -> None:
        """Test __call__ with single pipeline."""
        self.two_crop_transform.is_both = False

        results = self.two_crop_transform(inputs_np)

        assert isinstance(results, dict)
        assert "img" in results and results["img"].ndim == 3
        assert "gt_semantic_seg" in results and results["gt_semantic_seg"].ndim == 2
        assert "flip" in results and isinstance(results["flip"], bool)


@e2e_pytest_unit
def test_random_resized_crop(inputs_PIL: Dict[str, Any]) -> None:
    """Test RandomResizedCrop."""
    random_resized_crop = RandomResizedCrop(size=(8, 8))

    results = random_resized_crop(inputs_PIL)

    assert isinstance(results, dict)
    assert "img" in results and results["img"].size == (8, 8)
    assert "gt_semantic_seg" in results and results["gt_semantic_seg"].shape == (8, 8)
    assert "img_shape" in results
    assert "ori_shape" in results
    assert "scale_factor" in results


@e2e_pytest_unit
def test_random_color_jitter(inputs_PIL: Dict[str, Any]) -> None:
    """Test RandomColorJitter."""
    random_color_jitter = RandomColorJitter(p=1.0)

    results = random_color_jitter(inputs_PIL)

    assert isinstance(results, dict)
    assert "img" in results


@e2e_pytest_unit
def test_random_grayscale(inputs_PIL: Dict[str, Any]) -> None:
    """Test RandomGrayscale."""
    random_grayscale = RandomGrayscale()

    results = random_grayscale(inputs_PIL)

    assert isinstance(results, dict)
    assert "img" in results


@e2e_pytest_unit
def test_random_gaussian_blur(inputs_PIL: Dict[str, Any]) -> None:
    """Test RandomGaussianBlur."""
    random_gaussian_blur = RandomGaussianBlur(p=1.0, kernel_size=3)

    results = random_gaussian_blur(inputs_PIL)

    assert isinstance(results, dict)
    assert "img" in results


@e2e_pytest_unit
def test_random_solarization(inputs_np: Dict[str, Any]) -> None:
    """Test RandomSolarization."""
    random_solarization = RandomSolarization(p=1.0)

    results = random_solarization(inputs_np)

    assert isinstance(results, dict)
    assert "img" in results
    assert repr(random_solarization) == "RandomSolarization"


@e2e_pytest_unit
def test_nd_array_to_pil_image(inputs_np: Dict[str, Any]) -> None:
    """Test NDArrayToPILImage."""
    nd_array_to_pil_image = NDArrayToPILImage(keys=["img"])

    results = nd_array_to_pil_image(inputs_np)

    assert "img" in results
    assert isinstance(results["img"], Image.Image)
    assert repr(nd_array_to_pil_image) == "NDArrayToPILImage"


@e2e_pytest_unit
def test_pil_image_to_nd_array(inputs_PIL: Dict[str, Any]) -> None:
    """Test PILImageToNDArray."""
    pil_image_to_nd_array = PILImageToNDArray(keys=["img"])

    results = pil_image_to_nd_array(inputs_PIL)

    assert "img" in results
    assert isinstance(results["img"], np.ndarray)
    assert repr(pil_image_to_nd_array) == "PILImageToNDArray"



"""Unit-Test case for otx.core.data.manager."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import shutil
from typing import List
from tempfile import TemporaryDirectory

import datumaro as dm
import pytest

from otx.cli.manager.config_manager import TASK_TYPE_TO_SUPPORTED_FORMAT
from otx.core.data.manager.dataset_manager import DatasetManager
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.core.data.test_helpers import (
    generate_datumaro_dataset,
    generate_datumaro_dataset_item,
)

AVAILABLE_TASKS = ["classification", "detection", "segmentation"]
AVAILABLE_SUBSETS = ["train", "val"]
AVAILABLE_DATA_ROOTS = [
    "tests/assets/classification_dataset",
    "tests/assets/car_tree_bug",
    "tests/assets/cityscapes_dataset/dataset",
    "tests/assets/anomaly/hazelnut",
    "tests/assets/cvat_dataset/action_classification/train",
]

DATA_ROOTS2FORMAT = {
    "tests/assets/classification_dataset": "imagenet",
    "tests/assets/car_tree_bug": "coco",
    "tests/assets/cityscapes_dataset/dataset": "cityscapes",
}


class TestOTXDatasetManager:
    def setup(self) -> None:
        self.dataset = {}
        for subset in AVAILABLE_SUBSETS:
            self.dataset[subset] = {}
            for task in AVAILABLE_TASKS:
                self.dataset[subset][task] = generate_datumaro_dataset(subsets=[subset], task=task)

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", AVAILABLE_TASKS)
    @pytest.mark.parametrize("subset", AVAILABLE_SUBSETS)
    def test_get_train_dataset(self, task: List[str], subset: List[str]):
        if subset == "val":
            with pytest.raises(ValueError, match="Can't find training data."):
                DatasetManager.get_train_dataset(self.dataset[subset][task])
        else:
            train_dataset = DatasetManager.get_train_dataset(self.dataset[subset][task])
            assert isinstance(train_dataset, dm.DatasetSubset)

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", AVAILABLE_TASKS)
    @pytest.mark.parametrize("subset", AVAILABLE_SUBSETS)
    def test_get_val_dataset(self, task: List[str], subset: List[str]):
        if subset == "train":
            assert DatasetManager.get_val_dataset(self.dataset[subset][task]) is None
        else:
            val_dataset = DatasetManager.get_val_dataset(self.dataset[subset][task])
            assert isinstance(val_dataset, dm.DatasetSubset)

    @e2e_pytest_unit
    @pytest.mark.parametrize("data_root", AVAILABLE_DATA_ROOTS)
    def test_get_data_format(self, data_root: str):
        assert isinstance(DatasetManager.get_data_format(data_root), str)

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", AVAILABLE_TASKS)
    @pytest.mark.parametrize("subset", AVAILABLE_SUBSETS)
    def test_get_image_path(self, task, subset):
        random_data = DatasetManager.get_image_path(
            generate_datumaro_dataset_item(item_id="0", subset=subset, task=task)
        )
        assert random_data is None

        with TemporaryDirectory() as temp_dir:
            random_data = DatasetManager.get_image_path(
                generate_datumaro_dataset_item(item_id="0", subset=subset, task=task, temp_dir=temp_dir)
            )
            assert random_data is not None

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", AVAILABLE_TASKS)
    @pytest.mark.parametrize("subset", AVAILABLE_SUBSETS)
    def test_export_dataset(self, task, subset, tmp_dir_path):
        data_format = TASK_TYPE_TO_SUPPORTED_FORMAT[task.upper()][0]
        DatasetManager.export_dataset(self.dataset[subset][task], tmp_dir_path, data_format, save_media=False)
        shutil.rmtree(tmp_dir_path)

    @e2e_pytest_unit
    @pytest.mark.parametrize("data_root", AVAILABLE_DATA_ROOTS[:3])
    def test_import_dataset(self, data_root):
        data_format = DATA_ROOTS2FORMAT[data_root]
        assert DatasetManager.import_dataset(data_root, data_format=data_format) is not None

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", AVAILABLE_TASKS)
    @pytest.mark.parametrize("subset", AVAILABLE_SUBSETS)
    def test_auto_split(self, task, subset):
        dataset = DatasetManager.auto_split(
            task=task, dataset=self.dataset[subset][task], split_ratio=[("train", 0.8), ("val", 0.2)]
        )
        assert dataset is not None



# Copyright (C) 2018-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy
import json
import math
from math import ceil
from os import path as osp
from tempfile import TemporaryDirectory
from typing import Union

import pytest

from otx.hpo import hyperband
from otx.hpo.hpo_base import TrialStatus
from otx.hpo.hyperband import AshaTrial, Bracket, HyperBand, Rung
from tests.test_suite.e2e_test_system import e2e_pytest_component


@pytest.fixture
def good_trial_args():
    return {"trial_id": "name", "configuration": {"hp1": 1, "hp2": 1.2}, "train_environment": {"subset_ratio": 0.5}}


@pytest.fixture
def trial(good_trial_args):
    return AshaTrial(**good_trial_args)


@pytest.fixture
def good_rung_args():
    return {"resource": 10, "num_required_trial": 16, "reduction_factor": 2, "rung_idx": 0}


def register_scores_to_trial(trial, scores=[val for val in range(100)]):
    if len(trial.score) != 0:
        base_resource = max(trial.score.keys())
    else:
        base_resource = 0
    for idx, score in enumerate(scores):
        trial.register_score(score, base_resource + idx + 1)


@pytest.fixture
def rung(good_rung_args):
    return Rung(**good_rung_args)


@pytest.fixture
def good_bracket_args():
    hp_configs = [AshaTrial(i, {"hp1": 1, "hp2": 1.2}) for i in range(100)]
    return {
        "bracket_id": 0,
        "minimum_resource": 4,
        "maximum_resource": 64,
        "hyper_parameter_configurations": hp_configs,
        "reduction_factor": 2,
        "mode": "max",
        "asynchronous_sha": True,
    }


@pytest.fixture
def bracket(good_bracket_args):
    return Bracket(**good_bracket_args)


@pytest.fixture
def good_hyperband_args():
    with TemporaryDirectory() as tmp_dir:
        yield {
            "search_space": {
                "hp1": {"param_type": "uniform", "max": 100, "min": 10},
                "hp2": {"param_type": "qloguniform", "max": 1000, "min": 100, "step": 2, "log_base": 10},
            },
            "save_path": tmp_dir,
            "mode": "max",
            "num_workers": 1,
            "num_full_iterations": 64,
            "non_pure_train_ratio": 0.2,
            "full_dataset_size": 100,
            "metric": "mAP",
            "maximum_resource": 64,
            "minimum_resource": 1,
            "reduction_factor": 4,
            "asynchronous_sha": True,
            "asynchronous_bracket": True,
            "acceptable_additional_time_ratio": 1,
        }


@pytest.fixture
def hyper_band(good_hyperband_args):
    return HyperBand(**good_hyperband_args)


@e2e_pytest_component
@pytest.mark.parametrize("reduction_factor", [4, 100, 4000])
def test_check_reduction_factor_value(reduction_factor):
    hyperband._check_reduction_factor_value(reduction_factor)


@e2e_pytest_component
@pytest.mark.parametrize("reduction_factor", [-10, 1])
def test_check_reduction_factor_lesser_value(reduction_factor):
    with pytest.raises(ValueError):
        hyperband._check_reduction_factor_value(reduction_factor)


class TestAshaTrial:
    @e2e_pytest_component
    @pytest.mark.parametrize("rung_val", [0, 10])
    def teste_set_rung(self, trial, rung_val):
        trial.rung = rung_val

    @e2e_pytest_component
    @pytest.mark.parametrize("rung_val", [-10, -3])
    def test_set_negative_rung(self, trial, rung_val):
        with pytest.raises(ValueError):
            trial.rung = rung_val

    @e2e_pytest_component
    @pytest.mark.parametrize("bracket_val", [0, 10])
    def teste_set_bracket(self, trial, bracket_val):
        trial.bracket = bracket_val

    @e2e_pytest_component
    @pytest.mark.parametrize("bracket_val", [-10, -3])
    def test_set_negative_bracket(self, trial, bracket_val):
        with pytest.raises(ValueError):
            trial.bracket = bracket_val

    @e2e_pytest_component
    def test_save_results(self, trial, tmp_path):
        rung_idx = 3
        trial.rung = rung_idx
        register_scores_to_trial(trial)
        save_path = osp.join(tmp_path, "test")
        trial.save_results(save_path)

        with open(save_path, "r") as f:
            result = json.load(f)
        assert result["id"] == "name"
        assert result["configuration"]["hp1"] == 1
        assert result["configuration"]["hp2"] == 1.2
        assert result["train_environment"]["subset_ratio"] == 0.5
        assert result["rung"] == rung_idx
        for key, val in result["score"].items():
            assert int(key) - 1 == val


class TestRung:
    @e2e_pytest_component
    def test_init(self, good_rung_args):
        Rung(**good_rung_args)

    @e2e_pytest_component
    @pytest.mark.parametrize("resource", [-10, 0])
    def test_init_resource_nenative(self, good_rung_args, resource):
        wrong_trial_args = good_rung_args
        wrong_trial_args["resource"] = resource
        with pytest.raises(ValueError):
            Rung(**wrong_trial_args)

    @e2e_pytest_component
    @pytest.mark.parametrize("num_required_trial", [-10, 0])
    def test_init_num_required_trial(self, good_rung_args, num_required_trial):
        wrong_trial_args = good_rung_args
        wrong_trial_args["num_required_trial"] = num_required_trial
        with pytest.raises(ValueError):
            Rung(**wrong_trial_args)

    @e2e_pytest_component
    @pytest.mark.parametrize("reduction_factor", [-10, 0, 1])
    def test_init_wrong_reduction_factor(self, good_rung_args, reduction_factor):
        wrong_trial_args = good_rung_args
        wrong_trial_args["reduction_factor"] = reduction_factor
        with pytest.raises(ValueError):
            Rung(**wrong_trial_args)

    @e2e_pytest_component
    @pytest.mark.parametrize("rung_idx", [-10, -3])
    def test_init_wrong_rung_idx(self, good_rung_args, rung_idx):
        wrong_trial_args = good_rung_args
        wrong_trial_args["rung_idx"] = rung_idx
        with pytest.raises(ValueError):
            Rung(**wrong_trial_args)

    @e2e_pytest_component
    def test_add_new_trial(self, rung, good_trial_args):
        for _ in range(rung.num_required_trial):
            trial = AshaTrial(**good_trial_args)
            rung.add_new_trial(trial)
            assert trial.rung == rung.rung_idx
            assert trial.iteration == rung.resource
            assert trial.status == TrialStatus.READY

    @e2e_pytest_component
    def test_add_too_many_trials(self, rung, good_trial_args):
        with pytest.raises(RuntimeError):
            for _ in range(rung.num_required_trial + 1):
                trial = AshaTrial(**good_trial_args)
                rung.add_new_trial(trial)

    @e2e_pytest_component
    @pytest.mark.parametrize("mode", ["max", "min"])
    def test_get_best_trial(self, rung, good_trial_args, mode):
        for score in range(rung.num_required_trial):
            trial = AshaTrial(**good_trial_args)
            rung.add_new_trial(trial)
            trial.register_score(score=score, resource=trial.iteration)

        best_trial = rung.get_best_trial(mode)

        if mode == "max":
            assert best_trial.get_best_score(mode) == rung.num_required_trial - 1
        else:
            assert best_trial.get_best_score(mode) == 0

    @e2e_pytest_component
    def test_get_best_trial_with_not_started_trial(self, rung, good_trial_args):
        for score in range(rung.num_required_trial - 1):
            trial = AshaTrial(**good_trial_args)
            rung.add_new_trial(trial)
            trial.register_score(score=score, resource=trial.iteration)

        trial = AshaTrial(**good_trial_args)
        rung.add_new_trial(trial)
        best_trial = rung.get_best_trial()

        assert best_trial.get_best_score() == rung.num_required_trial - 2

    @e2e_pytest_component
    def test_get_best_trial_when_best_trial_is_undone(self, rung, good_trial_args):
        for _ in range(rung.num_required_trial - 1):
            trial = AshaTrial(**good_trial_args)
            rung.add_new_trial(trial)
            trial.register_score(score=1, resource=trial.iteration)

        trial = AshaTrial(**good_trial_args)
        rung.add_new_trial(trial)
        trial.register_score(score=100, resource=1)
        best_trial = rung.get_best_trial()

        assert best_trial.get_best_score() == 100

    @e2e_pytest_component
    def test_get_best_trial_with_no_trial(self, rung):
        best_trial = rung.get_best_trial()
        assert best_trial is None

    @e2e_pytest_component
    def test_get_best_trial_wrong_mode_val(self, rung):
        with pytest.raises(ValueError):
            rung.get_best_trial("wrong")

    @e2e_pytest_component
    def test_need_more_trials(self, rung, good_trial_args):
        for _ in range(rung.num_required_trial):
            trial = AshaTrial(**good_trial_args)
            assert rung.need_more_trials()
            rung.add_new_trial(trial)

        assert not rung.need_more_trials()

    @e2e_pytest_component
    def test_get_num_trials(self, rung, good_trial_args):
        for idx in range(rung.num_required_trial):
            trial = AshaTrial(**good_trial_args)
            rung.add_new_trial(trial)
            assert rung.get_num_trials() == idx + 1

    @e2e_pytest_component
    def test_need_more_trails(self, rung, good_trial_args):
        for i in range(1, rung.num_required_trial + 1):
            trial = AshaTrial(**good_trial_args)
            rung.add_new_trial(trial)
            if i != rung.num_required_trial:
                assert rung.need_more_trials()
            else:
                assert not rung.need_more_trials()

    @e2e_pytest_component
    def test_is_done(self, rung, good_trial_args):
        for i in range(rung.num_required_trial - 1):
            trial = AshaTrial(**good_trial_args)
            register_scores_to_trial(trial, [val for val in range(rung.resource)])
            rung.add_new_trial(trial)
            assert not rung.is_done()

        trial = AshaTrial(**good_trial_args)
        register_scores_to_trial(trial, [val for val in range(rung.resource - 1)])
        rung.add_new_trial(trial)
        assert not rung.is_done()
        trial.register_score(100, rung.resource + 1)
        assert rung.is_done()

    @e2e_pytest_component
    def test_get_trial_to_promote_not_asha(self, rung, good_trial_args):
        maximum_score = 9999999
        for i in range(rung.num_required_trial - 1):
            trial = AshaTrial(**good_trial_args)
            register_scores_to_trial(trial, [val for val in range(rung.resource)])
            rung.add_new_trial(trial)

        assert rung.get_trial_to_promote() is None

        trial = AshaTrial(**good_trial_args)
        register_scores_to_trial(trial, [maximum_score for _ in range(rung.resource)])
        rung.add_new_trial(trial)
        assert rung.get_trial_to_promote() == trial

        num_promoteable = rung._num_required_trial // rung._reduction_factor
        for _ in range(num_promoteable - 1):
            best_trial = rung.get_trial_to_promote()
            best_trial.rung += 1
            assert rung.get_trial_to_promote(False) is not None

        best_trial = rung.get_trial_to_promote()
        best_trial.rung += 1
        assert rung.get_trial_to_promote(False) is None

    @e2e_pytest_component
    def test_get_trial_to_promote_asha(self, rung, good_trial_args):
        num_promoteable = rung._num_required_trial // rung._reduction_factor
        for i in range(num_promoteable // rung._reduction_factor):
            for _ in range(rung._reduction_factor):
                trial = AshaTrial(**good_trial_args)
                register_scores_to_trial(trial, [val for val in range(rung.resource)])
                rung.add_new_trial(trial)

            assert rung.get_trial_to_promote(True) is not None
            best_trial = rung.get_trial_to_promote(True)
            best_trial.rung += 1
            assert rung.get_trial_to_promote(True) is None

    @e2e_pytest_component
    def test_get_trial_to_promote_not_running(self, rung, good_trial_args):
        for i in range(rung.num_required_trial):
            trial = AshaTrial(**good_trial_args)
            rung.add_new_trial(trial)

        for i in range(rung.num_required_trial):
            trial = rung.get_next_trial()
            register_scores_to_trial(trial, [val for val in range(rung.resource)])
            trial.status = TrialStatus.RUNNING

        promoted_trial = rung.get_trial_to_promote()
        assert promoted_trial is None

        trial.status = TrialStatus.STOP
        promoted_trial = rung.get_trial_to_promote()
        assert promoted_trial.status != TrialStatus.RUNNING

    @e2e_pytest_component
    def test_get_next_trial(self, rung, good_trial_args):
        trial = AshaTrial(**good_trial_args)
        rung.add_new_trial(trial)

        new_trial = rung.get_next_trial()
        assert trial == new_trial
        trial.status = TrialStatus.RUNNING

        # running trial isn't provided
        new_trial = rung.get_next_trial()
        assert new_trial is None

        # finished trial isn't provided
        register_scores_to_trial(trial, [i for i in range(trial.iteration)])
        trial.status = TrialStatus.STOP
        new_trial = rung.get_next_trial()
        assert new_trial is None

    @e2e_pytest_component
    def test_get_next_trial_stopped_in_progress(self, rung, trial):
        rung.add_new_trial(trial)
        register_scores_to_trial(trial, [i for i in range(trial.iteration - 1)])
        undone_trial = rung.get_next_trial()
        assert trial == undone_trial


class TestBracket:
    @e2e_pytest_component
    def test_init(self, good_bracket_args):
        Bracket(**good_bracket_args)

    @e2e_pytest_component
    def test_init_minimum_is_negative(self, good_bracket_args):
        wrong_args = good_bracket_args
        wrong_args["minimum_resource"] = -1
        with pytest.raises(ValueError):
            Bracket(**wrong_args)

    @e2e_pytest_component
    @pytest.mark.parametrize("reduction_factor", [-10, 0, 1])
    def test_init_wrong_reduction_factor(self, good_bracket_args, reduction_factor):
        wrong_args = good_bracket_args
        wrong_args["reduction_factor"] = reduction_factor
        with pytest.raises(ValueError):
            Bracket(**wrong_args)

    @e2e_pytest_component
    def test_init_wrong_mode_val(self, good_bracket_args):
        wrong_args = good_bracket_args
        wrong_args["mode"] = "wrong"
        with pytest.raises(ValueError):
            Bracket(**wrong_args)

    @e2e_pytest_component
    def test_init_minimum_val_is_bigger_than_maximum_val(self, good_bracket_args):
        wrong_args = good_bracket_args
        wrong_args["minimum_resource"] = 100
        wrong_args["maximum_resource"] = 10
        with pytest.raises(ValueError):
            Bracket(**wrong_args)

    @e2e_pytest_component
    def test_init_empty_hyper_parameter_configurations(self, good_bracket_args):
        wrong_args = good_bracket_args
        wrong_args["hyper_parameter_configurations"] = []
        with pytest.raises(ValueError):
            Bracket(**wrong_args)

    @e2e_pytest_component
    def test_max_rung(self, good_bracket_args):
        bracket = Bracket(**good_bracket_args)
        expected_val = math.ceil(
            math.log(
                good_bracket_args["maximum_resource"] / good_bracket_args["minimum_resource"],
                good_bracket_args["reduction_factor"],
            )
        )
        assert bracket.max_rung == expected_val

    @e2e_pytest_component
    def test_calcuate_max_rung_idx(self):
        minimum_resource = 1
        maximum_resource = 100
        reduction_factor = 3

        expected_val = math.ceil(math.log(maximum_resource / minimum_resource, reduction_factor))
        assert Bracket.calcuate_max_rung_idx(minimum_resource, maximum_resource, reduction_factor) == expected_val

    @e2e_pytest_component
    @pytest.mark.parametrize(
        "minimum_resource,maximum_resource,reduction_factor", [(-1, 100, 3), (1, -3, 3), (1, 100, -2), (10, 3, 3)]
    )
    def test_calcuate_max_rung_with_wrong_input(self, minimum_resource, maximum_resource, reduction_factor):
        with pytest.raises(ValueError):
            Bracket.calcuate_max_rung_idx(minimum_resource, maximum_resource, reduction_factor)

    @e2e_pytest_component
    def test_release_new_trial(self, bracket):
        while True:
            new_trial = bracket.get_next_trial()
            if new_trial is None:
                break
            new_trial.status = TrialStatus.RUNNING
            assert new_trial.bracket == bracket.id
            assert new_trial.rung == 0

    @e2e_pytest_component
    def test_promote_trial_if_available_asha(self, good_bracket_args):
        reduction_factor = good_bracket_args["reduction_factor"]
        bracket = Bracket(**good_bracket_args)
        for _ in range(reduction_factor):
            trial = bracket.get_next_trial()
            assert trial.rung == 0
            trial.register_score(score=0, resource=trial.iteration)

        trial = bracket.get_next_trial()
        assert trial.rung == 1

    @e2e_pytest_component
    def test_promote_trial_if_available_sha(self, good_bracket_args):
        good_bracket_args["asynchronous_sha"] = False
        bracket = Bracket(**good_bracket_args)
        trial_arr = []
        while True:
            trial = bracket.get_next_trial()
            if trial is None:
                break

            assert trial.rung == 0
            trial.status = TrialStatus.RUNNING
            trial_arr.append(trial)

        for trial in trial_arr:
            trial.register_score(score=1, resource=trial.iteration)
            trial.status = TrialStatus.READY

        trial = bracket.get_next_trial()
        assert trial.rung == 1

    @e2e_pytest_component
    def test_get_next_trial(self, bracket):
        while not bracket.is_done():
            trial = bracket.get_next_trial()
            assert trial is not None

            trial.status = TrialStatus.RUNNING
            trial.register_score(0, trial.iteration)
            trial.status = TrialStatus.STOP

        assert bracket.is_done()

    @e2e_pytest_component
    def test_get_next_trial_if_trial_is_always_running(self, bracket):
        trial_arr = []
        while True:
            trial = bracket.get_next_trial()
            if trial is None:
                break

            trial.status = TrialStatus.RUNNING
            trial_arr.append(trial)

        for trial in trial_arr:
            trial.register_score(0, trial.iteration)

        trial = bracket.get_next_trial()
        assert trial is None

    @e2e_pytest_component
    def test_is_done(self, bracket):
        while True:
            trial = bracket.get_next_trial()
            if trial is None:
                break

            trial.register_score(score=0, resource=trial.iteration)

        assert bracket.is_done()

    @e2e_pytest_component
    @pytest.mark.parametrize("num", [1, 5, 15])
    def test_num_trial_is_not_enough(self, good_bracket_args, num):
        wrong_bracket_args = good_bracket_args
        wrong_bracket_args["hyper_parameter_configurations"] = wrong_bracket_args["hyper_parameter_configurations"][
            :num
        ]

        with pytest.raises(ValueError):
            Bracket(**wrong_bracket_args)

    @e2e_pytest_component
    def test_get_best_trial(self, bracket):
        expected_score = 999999
        trial = bracket.get_next_trial()
        expected_trial_id = trial.id
        register_scores_to_trial(
            trial, [expected_score for _ in range(bracket._rungs[trial.rung].resource - trial.get_progress())]
        )
        while True:
            trial = bracket.get_next_trial()
            if trial is None:
                break

            register_scores_to_trial(
                trial, [score for score in range(bracket._rungs[trial.rung].resource - trial.get_progress())]
            )
        trial = bracket.get_best_trial()
        assert trial.get_best_score(bracket._mode) == expected_score
        assert trial.id == expected_trial_id

    @e2e_pytest_component
    def test_get_best_trial_given_absent_trial(self, bracket):
        assert bracket.get_best_trial() is None

    @e2e_pytest_component
    def test_get_best_trial_with_one_unfinished_trial(self, bracket):
        trial = bracket.get_next_trial()
        register_scores_to_trial(trial, [1])
        best_trial = bracket.get_best_trial()
        assert trial == best_trial

    @e2e_pytest_component
    def test_save_results(self, good_bracket_args, tmp_path):
        trial_num = len(good_bracket_args["hyper_parameter_configurations"])
        bracket = Bracket(**good_bracket_args)
        while True:
            trial = bracket.get_next_trial()
            if trial is None:
                break

            register_scores_to_trial(
                trial, [score for score in range(bracket._rungs[trial.rung].resource - trial.get_progress())]
            )

        bracket.save_results(tmp_path)

        with open(osp.join(tmp_path, "rung_status.json"), "r") as f:
            result = json.load(f)

        assert result["minimum_resource"] == good_bracket_args["minimum_resource"]
        assert result["maximum_resource"] == good_bracket_args["maximum_resource"]
        assert result["reduction_factor"] == good_bracket_args["reduction_factor"]
        assert result["mode"] == good_bracket_args["mode"]
        assert result["asynchronous_sha"] == good_bracket_args["asynchronous_sha"]
        assert result["num_trials"] == trial_num
        assert len(result["rung_status"]) == bracket.max_rung + 1
        for rung_status in result["rung_status"]:
            assert rung_status["num_trial"] == rung_status["num_required_trial"]
        for i in range(trial_num):
            assert osp.exists(osp.join(tmp_path, f"{i}.json"))

    @e2e_pytest_component
    def test_print_result(self, bracket):
        while True:
            trial = bracket.get_next_trial()
            if trial is None:
                break

            register_scores_to_trial(
                trial, [score for score in range(bracket._rungs[trial.rung].resource - trial.get_progress())]
            )

        bracket.print_result()

    @e2e_pytest_component
    def test_print_result_without_train(self, bracket):
        bracket.print_result()

    @e2e_pytest_component
    def test_report_trial_exit_abnormally(self, bracket):
        trial = bracket.get_next_trial()
        trial.register_score(score=0, resource=trial.iteration - 0.1)
        new_trial = bracket.get_next_trial()
        assert trial.id == new_trial.id


class TestHyperBand:
    @e2e_pytest_component
    def test_init(self, good_hyperband_args):
        HyperBand(**good_hyperband_args)

    @e2e_pytest_component
    @pytest.mark.parametrize("minimum_resource", [-10, 0])
    def test_init_not_postive_maximum_resource(self, good_hyperband_args, minimum_resource):
        wrong_arg = good_hyperband_args
        wrong_arg["minimum_resource"] = minimum_resource
        with pytest.raises(ValueError):
            HyperBand(**wrong_arg)

    @e2e_pytest_component
    @pytest.mark.parametrize("reduction_factor", [-10, 0, 1])
    def test_init_wrong_reduction_factor(self, good_hyperband_args, reduction_factor):
        wrong_arg = good_hyperband_args
        wrong_arg["reduction_factor"] = reduction_factor
        with pytest.raises(ValueError):
            HyperBand(**wrong_arg)

    @e2e_pytest_component
    def test_init_maximum_is_same_with_minimum(self, good_hyperband_args):
        good_hyperband_args["maximum_resource"] = good_hyperband_args["minimum_resource"]
        HyperBand(**good_hyperband_args)

    @e2e_pytest_component
    def test_init_no_minimum_resource(self, good_hyperband_args):
        del good_hyperband_args["minimum_resource"]
        HyperBand(**good_hyperband_args)

    @e2e_pytest_component
    def test_init_no_maximum_resource(self, good_hyperband_args):
        del good_hyperband_args["maximum_resource"]
        HyperBand(**good_hyperband_args)

    @e2e_pytest_component
    @pytest.mark.parametrize("num", [1, 10])
    def test_make_new_hyper_parameter_configs(self, good_hyperband_args, num):
        hb = HyperBand(**good_hyperband_args)

        while True:
            trial = hb.get_next_sample()
            if trial is None:
                break
            trial.status = TrialStatus.RUNNING

            assert 10 <= trial.configuration["hp1"] <= 100
            assert 100 <= trial.configuration["hp2"] <= 1000
            assert trial.configuration["hp2"] % 2 == 0

    @e2e_pytest_component
    def test_get_next_sample(self, hyper_band):
        while True:
            trial = hyper_band.get_next_sample()
            if trial is None:
                break
            hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=False)
            hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=True)

        assert hyper_band.is_done()

    @e2e_pytest_component
    def test_get_next_sample_without_train(self, hyper_band):
        while True:
            trial = hyper_band.get_next_sample()
            if trial is None:
                break
            trial.status = TrialStatus.RUNNING

        assert not hyper_band.is_done()

    @e2e_pytest_component
    def test_report_score(self, hyper_band):
        trial = hyper_band.get_next_sample()
        score = 100
        resource = 10
        hyper_band.report_score(score, resource, trial.id)
        assert trial.score[resource] == score

    @e2e_pytest_component
    def test_report_score_trial_done(self, hyper_band):
        trial = hyper_band.get_next_sample()
        hyper_band.report_score(100, 0.1, trial.id)
        hyper_band.report_score(0, 0, trial.id, done=True)
        assert trial.get_progress() == trial.iteration

    @e2e_pytest_component
    def test_get_best_config(self, hyper_band):
        max_score = 9999999
        trial = hyper_band.get_next_sample()
        expected_configuration = {"id": trial.id, "config": trial.configuration}
        hyper_band.report_score(score=max_score, resource=trial.iteration, trial_id=trial.id, done=False)
        hyper_band.report_score(score=max_score, resource=trial.iteration, trial_id=trial.id, done=True)
        while True:
            trial = hyper_band.get_next_sample()
            if trial is None:
                break
            hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=False)
            hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=True)

        best_config = hyper_band.get_best_config()

        assert best_config == expected_configuration

    @e2e_pytest_component
    def test_get_best_config_before_train(self, hyper_band):
        best_config = hyper_band.get_best_config()
        assert best_config is None

    @e2e_pytest_component
    def test_train_option_exists(self, hyper_band):
        trial = hyper_band.get_next_sample()
        train_config = trial.get_train_configuration()
        assert "subset_ratio" in train_config["train_environment"]

    @e2e_pytest_component
    def test_prior_hyper_parameters(self, good_hyperband_args):
        prior1 = {"hp1": 1, "hp2": 2}
        prior2 = {"hp1": 100, "hp2": 200}
        good_hyperband_args["prior_hyper_parameters"] = [prior1, prior2]
        hyper_band = HyperBand(**good_hyperband_args)
        first_trial = hyper_band.get_next_sample()
        first_trial.status = TrialStatus.RUNNING
        second_trial = hyper_band.get_next_sample()
        second_trial.status = TrialStatus.RUNNING

        print(first_trial.configuration)
        print(second_trial.configuration)

        assert first_trial.configuration == prior1
        assert second_trial.configuration == prior2

    @e2e_pytest_component
    @pytest.mark.parametrize("num_prior_param", [10, 100, 1000])
    def test_many_prior_hyper_parameters(self, good_hyperband_args, num_prior_param):
        prior_hyper_parameters = []
        for i in range(num_prior_param):
            new_prior = {}
            for key, val in good_hyperband_args["search_space"].items():
                new_prior[key] = val["min"] + i
            prior_hyper_parameters.append(new_prior)
        good_hyperband_args["prior_hyper_parameters"] = copy.deepcopy(prior_hyper_parameters)

        hyper_band = HyperBand(**good_hyperband_args)

        i = 0
        while i < num_prior_param:
            trial = hyper_band.get_next_sample()
            if trial is None:
                break

            if not trial.score:
                assert trial.configuration == prior_hyper_parameters[i]
                i += 1
            hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=False)
            hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=True)

        assert i == num_prior_param or hyper_band.is_done()

    @e2e_pytest_component
    def test_auto_config_decrease(self, good_hyperband_args):
        full_train_resource = good_hyperband_args["maximum_resource"]
        expected_time_ratio = 4
        good_hyperband_args["expected_time_ratio"] = expected_time_ratio
        hyperband = HyperBand(**good_hyperband_args)

        total_resource = 0
        while True:
            trial = hyperband.get_next_sample()
            if trial is None:
                break

            resource = ceil(trial.iteration - trial.get_progress())
            total_resource += resource

            hyperband.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=False)
            hyperband.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=True)

        maximum_resource = full_train_resource * expected_time_ratio * hyperband.acceptable_additional_time_ratio
        assert maximum_resource >= total_resource >= maximum_resource * 0.8

    @e2e_pytest_component
    def test_auto_config_increase(self, good_hyperband_args):
        full_train_resource = good_hyperband_args["maximum_resource"]
        expected_time_ratio = 100
        good_hyperband_args["expected_time_ratio"] = expected_time_ratio
        hyperband = HyperBand(**good_hyperband_args)

        total_resource = 0
        while True:
            trial = hyperband.get_next_sample()
            if trial is None:
                break

            resource = ceil(trial.iteration - trial.get_progress())
            total_resource += resource

            hyperband.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=False)
            hyperband.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=True)

        maximum_resource = full_train_resource * expected_time_ratio * hyperband.acceptable_additional_time_ratio
        assert maximum_resource >= total_resource >= maximum_resource * 0.8

    @e2e_pytest_component
    def test_asynchronous_bracket(self, hyper_band):
        bracket_id_arr = []
        while True:
            new_trial = hyper_band.get_next_sample()
            if new_trial is None:
                break

            new_trial.status = TrialStatus.RUNNING
            if new_trial.bracket not in bracket_id_arr:
                bracket_id_arr.append(new_trial.bracket)

        assert len(bracket_id_arr) > 1

    @e2e_pytest_component
    def test_synchronous_bracket(self, good_hyperband_args):
        good_hyperband_args["asynchronous_bracket"] = False
        hyper_band = HyperBand(**good_hyperband_args)
        bracket_id_arr = []
        while True:
            new_trial = hyper_band.get_next_sample()
            if new_trial is None:
                break

            new_trial.status = TrialStatus.RUNNING
            if new_trial.bracket not in bracket_id_arr:
                bracket_id_arr.append(new_trial.bracket)

        assert len(bracket_id_arr) == 1

    @e2e_pytest_component
    def test_print_result(self, hyper_band):
        while not hyper_band.is_done():
            trial = hyper_band.get_next_sample()
            if trial is None:
                break

            hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=False)
            hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=True)

        hyper_band.print_result()

    @e2e_pytest_component
    def test_print_result_without_train(self, hyper_band):
        hyper_band.print_result()

    @e2e_pytest_component
    def test_report_trial_exit_abnormally(self, hyper_band):
        trial = hyper_band.get_next_sample()
        hyper_band.report_score(score=50, resource=trial.iteration - 0.1, trial_id=trial.id, done=False)
        new_trial = hyper_band.get_next_sample()
        assert trial.id == new_trial.id

    @e2e_pytest_component
    def test_absence_minimum_resource(self, good_hyperband_args):
        del good_hyperband_args["minimum_resource"]
        hyper_band = HyperBand(**good_hyperband_args)
        first_trial = hyper_band.get_next_sample()
        assert first_trial.iteration == hyper_band.maximum_resource
        first_validation = 2
        hyper_band.report_score(score=1, resource=first_validation, trial_id=first_trial.id)

        iter_set = set()
        while not hyper_band.is_done():
            trial = hyper_band.get_next_sample()
            iter_set.add(trial.iteration)
            hyper_band.report_score(score=1, resource=trial.iteration, trial_id=trial.id)

        s_max = math.floor(
            math.log(hyper_band.maximum_resource / first_validation, good_hyperband_args["reduction_factor"])
        )
        expected_min = hyper_band.maximum_resource * (good_hyperband_args["reduction_factor"] ** -s_max)

        assert min(iter_set) == expected_min

    @e2e_pytest_component
    @pytest.mark.parametrize("num_trial_to_estimate", [10, 30, 100])
    def test_without_maximum_resource(self, good_hyperband_args, num_trial_to_estimate):
        del good_hyperband_args["maximum_resource"]
        max_validation = 120
        hyper_band = HyperBand(**good_hyperband_args)

        trials_to_estimate = []
        for _ in range(num_trial_to_estimate):
            trials_to_estimate.append(hyper_band.get_next_sample())

        for trial in reversed(trials_to_estimate[1:]):
            assert trial.iteration == good_hyperband_args["minimum_resource"]
            for iter in range(1, trial.iteration + 1):
                if hyper_band.report_score(score=1, resource=iter, trial_id=trial.id) == TrialStatus.STOP:
                    break
            assert iter == good_hyperband_args["minimum_resource"]

        first_trial = trials_to_estimate[0]
        hyper_band.report_score(score=1, resource=max_validation, trial_id=first_trial.id)
        hyper_band.report_score(score=0, resource=0, trial_id=first_trial.id, done=True)

        assert hyper_band.maximum_resource == max_validation

    @e2e_pytest_component
    @pytest.mark.parametrize("num_trial_to_estimate", [10, 30, 100])
    def test_auto_config_decrease_without_maximum_resource(self, good_hyperband_args, num_trial_to_estimate):
        """
        Validate auto config decreases ASHA resource well without maximum_resource.
        Current auto config doesn't consider already uesd resource by unused trials to estimate.
        To align that, this test also doesn't add it to total resource.
        """
        del good_hyperband_args["maximum_resource"]
        expected_time_ratio = 4
        max_validation = 120
        total_resource = max_validation
        good_hyperband_args["expected_time_ratio"] = expected_time_ratio
        hyperband = HyperBand(**good_hyperband_args)
        trial_ids = set()

        trials_to_estimate = []
        for _ in range(num_trial_to_estimate):
            trial = hyperband.get_next_sample()
            trials_to_estimate.append(trial)

        for trial in reversed(trials_to_estimate[1:]):
            assert trial.iteration == good_hyperband_args["minimum_resource"]
            for iter in range(1, trial.iteration + 1):
                if hyperband.report_score(score=1, resource=iter, trial_id=trial.id) == TrialStatus.STOP:
                    break

        first_trial = trials_to_estimate[0]
        hyperband.report_score(score=1, resource=max_validation, trial_id=first_trial.id)
        hyperband.report_score(score=0, resource=0, trial_id=first_trial.id, done=True)

        while True:
            trial = hyperband.get_next_sample()
            if trial is None:
                break
            if trial.id not in trial_ids:
                resource = trial.iteration
                trial_ids.add(trial.id)
            else:
                resource = trial.iteration - trial.get_progress()

            total_resource += resource

            hyperband.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=False)
            hyperband.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=True)

        maximum_resource = (
            hyperband.num_full_iterations * expected_time_ratio * hyperband.acceptable_additional_time_ratio
        )

        assert maximum_resource >= total_resource >= maximum_resource * 0.8

    @e2e_pytest_component
    @pytest.mark.parametrize("num_trial_to_estimate", [10, 30, 100])
    def test_auto_config_increase_without_maximum_resource(self, good_hyperband_args, num_trial_to_estimate):
        del good_hyperband_args["maximum_resource"]
        expected_time_ratio = 100
        max_validation = 120
        good_hyperband_args["expected_time_ratio"] = expected_time_ratio
        hyperband = HyperBand(**good_hyperband_args)

        trials_to_estimate = []
        for _ in range(num_trial_to_estimate):
            trial = hyperband.get_next_sample()
            trials_to_estimate.append(trial)

        for trial in reversed(trials_to_estimate[1:]):
            assert trial.iteration == good_hyperband_args["minimum_resource"]
            for iter in range(1, trial.iteration + 1):
                if hyperband.report_score(score=1, resource=iter, trial_id=trial.id) == TrialStatus.STOP:
                    break

        first_trial = trials_to_estimate[0]
        hyperband.report_score(score=1, resource=max_validation, trial_id=first_trial.id)
        hyperband.report_score(score=0, resource=0, trial_id=first_trial.id, done=True)

        total_resource = max_validation
        trial_ids = set()
        while True:
            trial = hyperband.get_next_sample()
            if trial is None:
                break
            if trial.id not in trial_ids:
                resource = trial.iteration
                trial_ids.add(trial.id)
            else:
                resource = trial.iteration - trial.get_progress()

            total_resource += resource

            hyperband.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=False)
            hyperband.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=True)

        maximum_resource = (
            hyperband.num_full_iterations * expected_time_ratio * hyperband.acceptable_additional_time_ratio
        )
        assert maximum_resource >= total_resource >= maximum_resource * 0.8

    @e2e_pytest_component
    @pytest.mark.parametrize("num_trial_to_estimate", [10, 30, 100])
    def test_without_minimum_maximum_resource(self, good_hyperband_args, num_trial_to_estimate):
        del good_hyperband_args["minimum_resource"]
        del good_hyperband_args["maximum_resource"]
        hyper_band = HyperBand(**good_hyperband_args)
        first_trial = hyper_band.get_next_sample()
        validation_interval = 2
        max_validation = 120

        trials_to_estimate = []
        for _ in range(num_trial_to_estimate):
            trial = hyper_band.get_next_sample()
            trials_to_estimate.append(trial)

        for trial in trials_to_estimate[1:]:
            for iter in range(validation_interval, trial.iteration + 1, validation_interval):
                if hyper_band.report_score(score=1, resource=iter, trial_id=trial.id) == TrialStatus.STOP:
                    break

        first_trial = trials_to_estimate[0]
        hyper_band.report_score(score=1, resource=max_validation, trial_id=first_trial.id)
        hyper_band.report_score(score=0, resource=0, trial_id=first_trial.id, done=True)

        iter_set = set()
        while not hyper_band.is_done():
            trial = hyper_band.get_next_sample()
            iter_set.add(trial.iteration)
            hyper_band.report_score(score=1, resource=trial.iteration, trial_id=trial.id)

        s_max = math.floor(
            math.log(hyper_band.maximum_resource / validation_interval, good_hyperband_args["reduction_factor"])
        )

        expected_min = hyper_band.maximum_resource * (good_hyperband_args["reduction_factor"] ** -s_max)

        assert min(iter_set) == expected_min
        assert hyper_band.maximum_resource == max_validation

    @e2e_pytest_component
    @pytest.mark.parametrize("expected_time_ratio", [3, 4, 5, 6])
    def test_hyperband_without_minimum_resource(self, good_hyperband_args, expected_time_ratio):
        """
        validate that when there is no minimum_resource,
        first trial stops near iteration given after ASHA schedule is made
        """
        good_hyperband_args["expected_time_ratio"] = expected_time_ratio
        del good_hyperband_args["minimum_resource"]
        hyper_band = HyperBand(**good_hyperband_args)

        val_interval = 3
        trial = hyper_band.get_next_sample()

        for iter in range(val_interval, trial.iteration + 1, val_interval):
            score = iter + 1
            trial_status = hyper_band.report_score(score, iter, trial.id)
            if trial_status == TrialStatus.STOP:
                break

        hyper_band.report_score(score, iter, trial.id, True)
        assert trial.get_progress() < trial.iteration + val_interval

    @e2e_pytest_component
    def test_get_done_progress(self, hyper_band: HyperBand):
        while not hyper_band.is_done():
            trial = hyper_band.get_next_sample()
            if trial is None:
                break

            hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=False)
            hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=True)

        assert hyper_band.get_progress() == 1

    @e2e_pytest_component
    @pytest.mark.parametrize("expected_time_ratio", [3, 4, 5, 6])
    def test_get_progress_with_expected_time_ratio(self, good_hyperband_args, expected_time_ratio):
        good_hyperband_args["expected_time_ratio"] = expected_time_ratio
        del good_hyperband_args["minimum_resource"]
        hyper_band = HyperBand(**good_hyperband_args)

        trial = hyper_band.get_next_sample()
        hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=False)
        hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=True)

        expected_total_resource = (
            good_hyperband_args["num_full_iterations"]
            * good_hyperband_args["num_workers"]
            * good_hyperband_args["acceptable_additional_time_ratio"]
            * expected_time_ratio
        )

        assert math.isclose(hyper_band.get_progress(), trial.get_progress() / expected_total_resource)

    @e2e_pytest_component
    def test_get_progress_with_out_expected_time_ratio(self, good_hyperband_args):
        hyper_band = HyperBand(**good_hyperband_args)
        full_asha_resource = _get_full_asha_resource(
            good_hyperband_args["maximum_resource"],
            good_hyperband_args["minimum_resource"],
            good_hyperband_args["reduction_factor"],
        )

        trial = hyper_band.get_next_sample()
        hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=False)
        hyper_band.report_score(score=50, resource=trial.iteration, trial_id=trial.id, done=True)

        assert math.isclose(hyper_band.get_progress(), trial.get_progress() / full_asha_resource)


def _get_full_asha_resource(
    maximum_resource: Union[float, int], minimum_resource: Union[float, int], reduction_factor: int
) -> Union[int, float]:
    total_resource: Union[int, float] = 0
    s_max = math.floor(math.log(maximum_resource / minimum_resource, reduction_factor))
    for idx in range(s_max + 1):
        num_max_rung_trials = math.floor((s_max + 1) / (idx + 1))
        total_resource += _calculate_bracket_resource(maximum_resource, reduction_factor, num_max_rung_trials, idx)

    return total_resource


def _calculate_bracket_resource(
    maximum_resource: Union[float, int],
    reduction_factor: Union[float, int],
    num_max_rung_trials: int,
    bracket_index: int,
) -> Union[int, float]:
    """Calculate how much resource is needed for the bracket given that resume is available."""
    num_trial = num_max_rung_trials * (reduction_factor**bracket_index)
    minimum_resource = maximum_resource * (reduction_factor**-bracket_index)

    total_resource = 0
    num_rungs = math.ceil(math.log(maximum_resource / minimum_resource, reduction_factor)) + 1
    previous_resource = 0
    resource = minimum_resource
    for _ in range(num_rungs):
        total_resource += num_trial * (resource - previous_resource)
        num_trial //= reduction_factor
        previous_resource = resource
        resource *= reduction_factor

    return total_resource



"""Code modified from: https://github.com/Atze00/MoViNet-pytorch/blob/main/movinets/models.py."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
from collections import OrderedDict
from typing import Any, Callable, Optional, Tuple, Union

import torch
import torch.nn.functional as F
from einops import rearrange
from mmaction.models.builder import BACKBONES
from mmcv.utils import Config
from torch import Tensor, nn
from torch.nn.modules.utils import _pair, _triple


class Conv2dBNActivation(nn.Sequential):
    """A base module that applies a 2D Conv-BN-Activation.

    Args:
        in_planes (int): Number of input channels.
        out_planes (int): Number of output channels.
        kernel_size (Union[int, Tuple[int, int]]): Size of the convolution kernel.
        padding (Union[int, Tuple[int, int]]): Size of the padding applied to the input.
        stride (Union[int, Tuple[int, int]], optional): Stride of the convolution. Default: 1.
        groups (int, optional): Number of groups in the convolution. Default: 1.
        norm_layer (Optional[Callable[..., nn.Module]], optional): Normalization layer to use.
            If None, identity is used. Default: None.
        activation_layer (Optional[Callable[..., nn.Module]], optional): Activation layer to use.
            If None, identity is used. Default: None.
        **kwargs (Any): Additional keyword arguments passed to nn.Conv2d.

    Attributes:
        kernel_size (Tuple[int, int]): Size of the convolution kernel.
        stride (Tuple[int, int]): Stride of the convolution.
        out_channels (int): Number of output channels.

    """

    def __init__(
        self,
        in_planes: int,
        out_planes: int,
        *,
        kernel_size: Union[int, Tuple[int, int]],
        padding: Union[int, Tuple[int, int]],
        stride: Union[int, Tuple[int, int]] = 1,
        groups: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        activation_layer: Optional[Callable[..., nn.Module]] = None,
        **kwargs: Any,
    ) -> None:
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        if norm_layer is None:
            norm_layer = nn.Identity
        if activation_layer is None:
            activation_layer = nn.Identity
        self.kernel_size = kernel_size
        self.stride = stride
        dict_layers = OrderedDict(
            {
                "conv2d": nn.Conv2d(
                    in_planes,
                    out_planes,
                    kernel_size=kernel_size,
                    stride=stride,
                    padding=padding,
                    groups=groups,
                    **kwargs,
                ),
                "norm": norm_layer(out_planes, eps=0.001),
                "act": activation_layer(),
            }
        )

        self.out_channels = out_planes
        super().__init__(dict_layers)


class Conv3DBNActivation(nn.Sequential):
    """A base module that applies a 3D Conv-BN-Activation.

    Args:
        in_planes (int): Number of input channels.
        out_planes (int): Number of output channels.
        kernel_size (Union[int, Tuple[int, int, int]]): Size of the convolution kernel.
        padding (Union[int, Tuple[int, int, int]]): Size of the padding applied to the input.
        stride (Union[int, Tuple[int, int, int]], optional): Stride of the convolution. Default: 1.
        groups (int, optional): Number of groups in the convolution. Default: 1.
        norm_layer (Optional[Callable[..., nn.Module]], optional): Normalization layer to use.
            If None, identity is used. Default: None.
        activation_layer (Optional[Callable[..., nn.Module]], optional): Activation layer to use.
            If None, identity is used. Default: None.
        **kwargs (Any): Additional keyword arguments passed to nn.Conv3d.

    Attributes:
        kernel_size (Tuple[int, int, int]): Size of the convolution kernel.
        stride (Tuple[int, int, int]): Stride of the convolution.
        out_channels (int): Number of output channels.

    """

    def __init__(
        self,
        in_planes: int,
        out_planes: int,
        *,
        kernel_size: Union[int, Tuple[int, int, int]],
        padding: Union[int, Tuple[int, int, int]],
        stride: Union[int, Tuple[int, int, int]] = 1,
        groups: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        activation_layer: Optional[Callable[..., nn.Module]] = None,
        **kwargs: Any,
    ) -> None:
        kernel_size = _triple(kernel_size)
        stride = _triple(stride)
        padding = _triple(padding)
        if norm_layer is None:
            norm_layer = nn.Identity
        if activation_layer is None:
            activation_layer = nn.Identity
        self.kernel_size = kernel_size
        self.stride = stride

        dict_layers = OrderedDict(
            {
                "conv3d": nn.Conv3d(
                    in_planes,
                    out_planes,
                    kernel_size=kernel_size,
                    stride=stride,
                    padding=padding,
                    groups=groups,
                    **kwargs,
                ),
                "norm": norm_layer(out_planes, eps=0.001),
                "act": activation_layer(),
            }
        )

        self.out_channels = out_planes
        super().__init__(dict_layers)


class ConvBlock3D(nn.Module):
    """A module that applies a 2+1D or 3D Conv-BN-activation sequential.

    Args:
        in_planes (int): Number of input channels.
        out_planes (int): Number of output channels.
        kernel_size (Tuple[int, int, int]): Size of the convolution kernel.
        tf_like (bool): Whether to use TensorFlow-like padding and convolution.
        conv_type (str): Type of 3D convolution to use. Must be "2plus1d" or "3d".
        padding (Tuple[int, int, int], optional): Size of the padding applied to the input.
            Default: (0, 0, 0).
        stride (Tuple[int, int, int], optional): Stride of the convolution. Default: (1, 1, 1).
        norm_layer (Optional[Callable[..., nn.Module]], optional): Normalization layer to use.
            If None, identity is used. Default: None.
        activation_layer (Optional[Callable[..., nn.Module]], optional): Activation layer to use.
            If None, identity is used. Default: None.
        bias (bool, optional): Whether to use bias in the convolution. Default: False.
        **kwargs (Any): Additional keyword arguments passed to nn.Conv2d or nn.Conv3d.

    Attributes:
        conv_1 (Union[Conv2dBNActivation, Conv3DBNActivation]): Convolutional layer.
        conv_2 (Optional[Conv2dBNActivation]): Convolutional layer for 2+1D convolution.
        padding (Tuple[int, int, int]): Size of the padding applied to the input.
        kernel_size (Tuple[int, int, int]): Size of the convolution kernel.
        dim_pad (int): Padding along the temporal dimension.
        stride (Tuple[int, int, int]): Stride of the convolution.
        conv_type (str): Type of 3D convolution used.
        tf_like (bool): Whether to use TensorFlow-like padding and convolution.

    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments
    def __init__(
        self,
        in_planes: int,
        out_planes: int,
        kernel_size: Tuple[int, int, int],
        tf_like: bool,
        conv_type: str,
        padding: Tuple[int, int, int] = (0, 0, 0),
        stride: Tuple[int, int, int] = (1, 1, 1),
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        activation_layer: Optional[Callable[..., nn.Module]] = None,
        bias: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.conv_2 = None
        if tf_like:
            # We need odd kernel to have even padding
            # and stride == 1 to precompute padding,
            if kernel_size[0] % 2 == 0:
                raise ValueError("tf_like supports only odd" + " kernels for temporal dimension")
            padding = ((kernel_size[0] - 1) // 2, 0, 0)
            if stride[0] != 1:
                raise ValueError("illegal stride value, tf like supports" + " only stride == 1 for temporal dimension")
            if stride[1] > kernel_size[1] or stride[2] > kernel_size[2]:
                # these values are not tested so should be avoided
                raise ValueError("tf_like supports only" + "  stride <= of the kernel size")

        if conv_type not in ["2plus1d", "3d"]:
            raise ValueError("only 2plus2d or 3d are " + "allowed as 3d convolutions")

        if conv_type == "2plus1d":
            self.conv_1 = Conv2dBNActivation(
                in_planes,
                out_planes,
                kernel_size=(kernel_size[1], kernel_size[2]),
                padding=(padding[1], padding[2]),
                stride=(stride[1], stride[2]),
                activation_layer=activation_layer,
                norm_layer=norm_layer,
                bias=bias,
                **kwargs,
            )
            if kernel_size[0] > 1:
                self.conv_2 = Conv2dBNActivation(
                    in_planes,
                    out_planes,
                    kernel_size=(kernel_size[0], 1),
                    padding=(padding[0], 0),
                    stride=(stride[0], 1),
                    activation_layer=activation_layer,
                    norm_layer=norm_layer,
                    bias=bias,
                    **kwargs,
                )
        elif conv_type == "3d":
            self.conv_1 = Conv3DBNActivation(
                in_planes,
                out_planes,
                kernel_size=kernel_size,
                padding=padding,
                activation_layer=activation_layer,
                norm_layer=norm_layer,
                stride=stride,
                bias=bias,
                **kwargs,
            )
        self.padding = padding
        self.kernel_size = kernel_size
        self.dim_pad = self.kernel_size[0] - 1
        self.stride = stride
        self.conv_type = conv_type
        self.tf_like = tf_like

    def _forward(self, x: Tensor) -> Tensor:
        shape_with_buffer = x.shape
        if self.conv_type == "2plus1d":
            x = rearrange(x, "b c t h w -> (b t) c h w")
        x = self.conv_1(x)
        if self.conv_type == "2plus1d":
            x = rearrange(x, "(b t) c h w -> b c t h w", t=shape_with_buffer[2])
            if self.conv_2 is not None:
                w = x.shape[-1]
                x = rearrange(x, "b c t h w -> b c t (h w)")
                x = self.conv_2(x)
                x = rearrange(x, "b c t (h w) -> b c t h w", w=w)
        return x

    def forward(self, x: Tensor) -> Tensor:
        """Forward function of ConvBlock3D."""
        if self.tf_like:
            x = same_padding(
                x,
                x.shape[-2],
                x.shape[-1],
                self.stride[-2],
                self.stride[-1],
                self.kernel_size[-2],
                self.kernel_size[-1],
            )
        x = self._forward(x)
        return x


class SqueezeExcitation(nn.Module):
    """Implements the Squeeze-and-Excitation (SE) block.

    Args:
        input_channels (int): Number of input channels.
        activation_2 (nn.Module): Activation function applied after the second convolutional block.
        activation_1 (nn.Module): Activation function applied after the first convolutional block.
        conv_type (str): Convolutional block type ("2plus1d" or "3d").
        squeeze_factor (int, optional): The reduction factor for the number of channels (default: 4).
        bias (bool, optional): Whether to add a bias term to the convolutional blocks (default: True).
    """

    def __init__(
        self,
        input_channels: int,
        activation_2: nn.Module,
        activation_1: nn.Module,
        conv_type: str,
        squeeze_factor: int = 4,
        bias: bool = True,
    ) -> None:
        super().__init__()
        se_multiplier = 1
        squeeze_channels = _make_divisible(input_channels // squeeze_factor * se_multiplier, 8)
        self.fc1 = ConvBlock3D(
            input_channels * se_multiplier,
            squeeze_channels,
            kernel_size=(1, 1, 1),
            padding=(0, 0, 0),
            tf_like=False,
            conv_type=conv_type,
            bias=bias,
        )
        self.activation_1 = activation_1()
        self.activation_2 = activation_2()
        self.fc2 = ConvBlock3D(
            squeeze_channels,
            input_channels,
            kernel_size=(1, 1, 1),
            padding=(0, 0, 0),
            tf_like=False,
            conv_type=conv_type,
            bias=bias,
        )

    def _scale(self, x: Tensor) -> Tensor:
        """Computes the scaling factor for the input tensor.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, channels, time, height, width).

        Returns:
            torch.Tensor: Scaling factor for the input tensor of shape (batch_size, channels, 1, 1, 1).
        """
        scale = F.adaptive_avg_pool3d(x, 1)
        scale = self.fc1(scale)
        scale = self.activation_1(scale)
        scale = self.fc2(scale)
        return self.activation_2(scale)

    def forward(self, x: Tensor) -> Tensor:
        """Forward function of SqueezeExcitation."""
        scale = self._scale(x)
        return scale * x


def _make_divisible(value: float, divisor: int, min_value: Optional[int] = None) -> int:
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(value + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * value:
        new_v += divisor
    return new_v


def same_padding(
    x: Tensor, in_height: int, in_width: int, stride_h: int, stride_w: int, filter_height: int, filter_width: int
) -> Tensor:
    """Applies padding to the input tensor to ensure that the output tensor size is the same as the input tensor size.

    Args:
        x (torch.Tensor): Input tensor of shape (batch_size, channels, time, height, width).
        in_height (int): Height of the input tensor.
        in_width (int): Width of the input tensor.
        stride_h (int): Stride in the height dimension.
        stride_w (int): Stride in the width dimension.
        filter_height (int): Height of the filter (kernel).
        filter_width (int): Width of the filter (kernel).

    Returns:
        torch.Tensor: Padded tensor of shape (batch_size, channels, time, height + pad_h, width + pad_w), where
        pad_h and pad_w are the heights and widths of the top, bottom, left, and right padding applied to the tensor.

    """
    if in_height % stride_h == 0:
        pad_along_height = max(filter_height - stride_h, 0)
    else:
        pad_along_height = max(filter_height - (in_height % stride_h), 0)
    if in_width % stride_w == 0:
        pad_along_width = max(filter_width - stride_w, 0)
    else:
        pad_along_width = max(filter_width - (in_width % stride_w), 0)
    pad_top = pad_along_height // 2
    pad_bottom = pad_along_height - pad_top
    pad_left = pad_along_width // 2
    pad_right = pad_along_width - pad_left
    padding_pad = (pad_left, pad_right, pad_top, pad_bottom)
    return torch.nn.functional.pad(x, padding_pad)


class TFAvgPool3D(nn.Module):
    """3D average pooling layer with padding."""

    def __init__(self) -> None:
        super().__init__()
        self.avgf = nn.AvgPool3d((1, 3, 3), stride=(1, 2, 2))

    def forward(self, x: Tensor) -> Tensor:
        """Applies 3D average pooling with padding to the input tensor.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, channels, time, height, width).

        Returns:
            torch.Tensor: Pooled tensor of shape (batch_size, channels, time, height', width'), where
            height' and width' are the heights and widths of the pooled tensor after padding is applied.

        """
        use_padding = x.shape[-1] % 2 != 0
        if use_padding:
            padding_pad = (0, 0, 0, 0)
        else:
            padding_pad = (0, 1, 0, 1)
        x = torch.nn.functional.pad(x, padding_pad)
        if use_padding:
            x = torch.nn.functional.avg_pool3d(
                x, (1, 3, 3), stride=(1, 2, 2), count_include_pad=False, padding=(0, 1, 1)
            )
        else:
            x = self.avgf(x)
            x[..., -1] = x[..., -1] * 9 / 6
            x[..., -1, :] = x[..., -1, :] * 9 / 6
        return x


class BasicBneck(nn.Module):
    """Basic bottleneck block of MoViNet network.

    Args:
        cfg (Config): Configuration object containing block's hyperparameters.
        tf_like (bool): A boolean indicating whether to use TensorFlow like convolution
            padding or not.
        conv_type (str): A string indicating the type of convolutional layer to use.
            Can be "2d" or "3d".
        norm_layer (Callable[..., nn.Module], optional): A callable normalization layer
            to use. Defaults to None.
        activation_layer (Callable[..., nn.Module], optional): A callable activation
            layer to use. Defaults to None.

    Attributes:
        expand (ConvBlock3D, optional): An optional expansion convolutional block.
        deep (ConvBlock3D): A convolutional block with kernel size, stride, padding,
            and groups as specified in the configuration object.
        squeeze_excitation (SqueezeExcitation): A squeeze-and-excitation block.
        project (ConvBlock3D): A projection convolutional block.
        res (nn.Sequential, optional): An optional residual convolutional block.
        alpha (nn.Parameter): A learnable parameter used in the ReZero operation.

    Raises:
        AssertionError: If the stride in configuration is not a tuple.

    """

    def __init__(
        self,
        cfg: "Config",
        tf_like: bool,
        conv_type: str,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        activation_layer: Optional[Callable[..., nn.Module]] = None,
    ) -> None:
        super().__init__()
        assert isinstance(cfg.stride, tuple)
        self.res = None

        layers = []
        if cfg.expanded_channels != cfg.out_channels:
            self.expand = ConvBlock3D(
                in_planes=cfg.input_channels,
                out_planes=cfg.expanded_channels,
                kernel_size=(1, 1, 1),
                padding=(0, 0, 0),
                conv_type=conv_type,
                tf_like=tf_like,
                norm_layer=norm_layer,
                activation_layer=activation_layer,
            )
        self.deep = ConvBlock3D(
            in_planes=cfg.expanded_channels,
            out_planes=cfg.expanded_channels,
            kernel_size=cfg.kernel_size,
            padding=cfg.padding,
            stride=cfg.stride,  # type: ignore
            groups=cfg.expanded_channels,
            conv_type=conv_type,
            tf_like=tf_like,
            norm_layer=norm_layer,
            activation_layer=activation_layer,
        )
        # pylint: disable=invalid-name
        self.se = SqueezeExcitation(
            cfg.expanded_channels,
            activation_1=activation_layer,
            activation_2=(nn.Sigmoid if conv_type == "3d" else nn.Hardsigmoid),
            conv_type=conv_type,
        )
        self.project = ConvBlock3D(
            cfg.expanded_channels,
            cfg.out_channels,
            kernel_size=(1, 1, 1),
            padding=(0, 0, 0),
            conv_type=conv_type,
            tf_like=tf_like,
            norm_layer=norm_layer,
            activation_layer=nn.Identity,
        )

        if not (cfg.stride == (1, 1, 1) and cfg.input_channels == cfg.out_channels):
            if cfg.stride != (1, 1, 1):
                if tf_like:
                    layers.append(TFAvgPool3D())
                else:
                    layers.append(nn.AvgPool3d((1, 3, 3), stride=cfg.stride, padding=cfg.padding_avg))
            layers.append(
                ConvBlock3D(
                    in_planes=cfg.input_channels,
                    out_planes=cfg.out_channels,
                    kernel_size=(1, 1, 1),
                    padding=(0, 0, 0),
                    norm_layer=norm_layer,
                    activation_layer=nn.Identity,
                    conv_type=conv_type,
                    tf_like=tf_like,
                )
            )
            self.res = nn.Sequential(*layers)
        # ReZero
        self.alpha = nn.Parameter(torch.tensor(0.0), requires_grad=True)

    def forward(self, x: Tensor) -> Tensor:
        """Forward function of BasicBneck."""
        if self.res is not None:
            residual = self.res(x)
        else:
            residual = x
        if hasattr(self, "expand"):
            x = self.expand(x)
        x = self.deep(x)
        x = self.se(x)
        x = self.project(x)
        result = residual + self.alpha * x
        return result


class MoViNet(nn.Module):
    """MoViNet class used for video classification.

    Args:
        cfg (Config): Configuration object containing network's hyperparameters.
        conv_type (str, optional): A string indicating the type of convolutional layer
            to use. Can be "2d" or "3d". Defaults to "3d".
        tf_like (bool, optional): A boolean indicating whether to use TensorFlow like
            convolution padding or not. Defaults to False.

    Attributes:
        conv1 (ConvBlock3D): A convolutional block for the first layer.
        blocks (nn.Sequential): A sequence of basic bottleneck blocks.
        conv7 (ConvBlock3D): A convolutional block for the final layer.

    Methods:
        avg(x: Tensor) -> Tensor: A static method that returns the adaptive average pool
            of the input tensor.
        _init_weights(module): A private method that initializes the weights of the network's
            convolutional, batch normalization, and linear layers.
        forward(x: Tensor) -> Tensor: The forward pass of the network.

    """

    def __init__(
        self,
        cfg: "Config",
        conv_type: str = "3d",
        tf_like: bool = False,
    ) -> None:
        super().__init__()
        tf_like = True
        blocks_dic = OrderedDict()

        norm_layer = nn.BatchNorm3d if conv_type == "3d" else nn.BatchNorm2d
        activation_layer = nn.SiLU if conv_type == "3d" else nn.Hardswish

        self.conv1 = ConvBlock3D(
            in_planes=cfg.conv1.input_channels,
            out_planes=cfg.conv1.out_channels,
            kernel_size=cfg.conv1.kernel_size,
            stride=cfg.conv1.stride,
            padding=cfg.conv1.padding,
            conv_type=conv_type,
            tf_like=tf_like,
            norm_layer=norm_layer,
            activation_layer=activation_layer,
        )
        for i, block in enumerate(cfg.blocks):
            for j, basicblock in enumerate(block):
                blocks_dic[f"b{i}_l{j}"] = BasicBneck(
                    basicblock,
                    conv_type=conv_type,
                    tf_like=tf_like,
                    norm_layer=norm_layer,
                    activation_layer=activation_layer,
                )
        self.blocks = nn.Sequential(blocks_dic)
        self.conv7 = ConvBlock3D(
            in_planes=cfg.conv7.input_channels,
            out_planes=cfg.conv7.out_channels,
            kernel_size=cfg.conv7.kernel_size,
            stride=cfg.conv7.stride,
            padding=cfg.conv7.padding,
            conv_type=conv_type,
            tf_like=tf_like,
            norm_layer=norm_layer,
            activation_layer=activation_layer,
        )

    def avg(self, x: Tensor) -> Tensor:
        """Returns the adaptive average pool of the input tensor.

        Args:
            x (Tensor): A tensor to be averaged.

        Returns:
            Tensor: A tensor with the averaged values.

        """
        return F.adaptive_avg_pool3d(x, 1)

    @staticmethod
    def _init_weights(module):
        if isinstance(module, nn.Conv3d):
            nn.init.kaiming_normal_(module.weight, mode="fan_out")
            if module.bias is not None:
                nn.init.zeros_(module.bias)
        elif isinstance(module, (nn.BatchNorm3d, nn.BatchNorm2d, nn.GroupNorm)):
            nn.init.ones_(module.weight)
            nn.init.zeros_(module.bias)
        elif isinstance(module, nn.Linear):
            nn.init.normal_(module.weight, 0, 0.01)
            nn.init.zeros_(module.bias)

    def forward(self, x: Tensor) -> Tensor:
        """Forward function of MoViNet."""
        x = self.conv1(x)
        x = self.blocks(x)
        x = self.conv7(x)
        x = self.avg(x)
        return x

    def init_weights(self):
        """Initializes the weights of network."""
        self.apply(self._init_weights)


@BACKBONES.register_module()
class OTXMoViNet(MoViNet):
    """MoViNet wrapper class for OTX."""

    # pylint: disable=unused-argument
    def __init__(self, **kwargs):
        cfg = Config()
        cfg.name = "A0"
        cfg.conv1 = Config()
        OTXMoViNet.fill_conv(cfg.conv1, 3, 8, (1, 3, 3), (1, 2, 2), (0, 1, 1))

        cfg.blocks = [
            [Config()],
            [Config() for _ in range(3)],
            [Config() for _ in range(3)],
            [Config() for _ in range(4)],
            [Config() for _ in range(4)],
        ]

        # block 2
        OTXMoViNet.fill_se_config(cfg.blocks[0][0], 8, 8, 24, (1, 5, 5), (1, 2, 2), (0, 2, 2), (0, 1, 1))

        # block 3
        OTXMoViNet.fill_se_config(cfg.blocks[1][0], 8, 32, 80, (3, 3, 3), (1, 2, 2), (1, 0, 0), (0, 0, 0))
        OTXMoViNet.fill_se_config(cfg.blocks[1][1], 32, 32, 80, (3, 3, 3), (1, 1, 1), (1, 1, 1), (0, 1, 1))
        OTXMoViNet.fill_se_config(cfg.blocks[1][2], 32, 32, 80, (3, 3, 3), (1, 1, 1), (1, 1, 1), (0, 1, 1))

        # block 4
        OTXMoViNet.fill_se_config(cfg.blocks[2][0], 32, 56, 184, (5, 3, 3), (1, 2, 2), (2, 0, 0), (0, 0, 0))
        OTXMoViNet.fill_se_config(cfg.blocks[2][1], 56, 56, 112, (3, 3, 3), (1, 1, 1), (1, 1, 1), (0, 1, 1))
        OTXMoViNet.fill_se_config(cfg.blocks[2][2], 56, 56, 184, (3, 3, 3), (1, 1, 1), (1, 1, 1), (0, 1, 1))

        # block 5
        OTXMoViNet.fill_se_config(cfg.blocks[3][0], 56, 56, 184, (5, 3, 3), (1, 1, 1), (2, 1, 1), (0, 1, 1))
        OTXMoViNet.fill_se_config(cfg.blocks[3][1], 56, 56, 184, (3, 3, 3), (1, 1, 1), (1, 1, 1), (0, 1, 1))
        OTXMoViNet.fill_se_config(cfg.blocks[3][2], 56, 56, 184, (3, 3, 3), (1, 1, 1), (1, 1, 1), (0, 1, 1))
        OTXMoViNet.fill_se_config(cfg.blocks[3][3], 56, 56, 184, (3, 3, 3), (1, 1, 1), (1, 1, 1), (0, 1, 1))

        # block 6
        OTXMoViNet.fill_se_config(cfg.blocks[4][0], 56, 104, 384, (5, 3, 3), (1, 2, 2), (2, 1, 1), (0, 1, 1))
        OTXMoViNet.fill_se_config(cfg.blocks[4][1], 104, 104, 280, (1, 5, 5), (1, 1, 1), (0, 2, 2), (0, 1, 1))
        OTXMoViNet.fill_se_config(cfg.blocks[4][2], 104, 104, 280, (1, 5, 5), (1, 1, 1), (0, 2, 2), (0, 1, 1))
        OTXMoViNet.fill_se_config(cfg.blocks[4][3], 104, 104, 344, (1, 5, 5), (1, 1, 1), (0, 2, 2), (0, 1, 1))

        cfg.conv7 = Config()
        OTXMoViNet.fill_conv(cfg.conv7, 104, 480, (1, 1, 1), (1, 1, 1), (0, 0, 0))

        cfg.dense9 = Config(dict(hidden_dim=2048))
        super().__init__(cfg)

    @staticmethod
    # pylint: disable=too-many-arguments
    def fill_se_config(
        conf,
        input_channels,
        out_channels,
        expanded_channels,
        kernel_size,
        stride,
        padding,
        padding_avg,
    ):
        """Set the values of a given Config object to SE module.

        Args:
            conf (Config): The Config object to be updated.
            input_channels (int): The number of input channels.
            out_channels (int): The number of output channels.
            expanded_channels (int): The number of channels after expansion in the basic block.
            kernel_size (tuple[int]): The size of the kernel.
            stride (tuple[int]): The stride of the kernel.
            padding (tuple[int]): The padding of the kernel.
            padding_avg (tuple[int]): The padding for the average pooling operation.

        Returns:
            None.
        """
        conf.expanded_channels = expanded_channels
        conf.padding_avg = padding_avg
        OTXMoViNet.fill_conv(
            conf,
            input_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
        )

    @staticmethod
    def fill_conv(
        conf,
        input_channels,
        out_channels,
        kernel_size,
        stride,
        padding,
    ):
        """Set the values of a given Config object to conv layer.

        Args:
            conf (Config): The Config object to be updated.
            input_channels (int): The number of input channels.
            out_channels (int): The number of output channels.
            kernel_size (tuple[int]): The size of the kernel.
            stride (tuple[int]): The stride of the kernel.
            padding (tuple[int]): The padding of the kernel.

        Returns:
            None.
        """
        conf.input_channels = input_channels
        conf.out_channels = out_channels
        conf.kernel_size = kernel_size
        conf.stride = stride
        conf.padding = padding



"""Custom MaskRCNN detector with tile classifier."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import numpy as np
import torch
from mmcv.cnn import ConvModule
from mmcv.runner import auto_fp16
from mmdet.models.builder import DETECTORS
from torch import nn

from otx.algorithms.common.adapters.mmdeploy import is_mmdeploy_enabled
from otx.algorithms.common.adapters.nncf import no_nncf_trace

from .custom_maskrcnn_detector import CustomMaskRCNN


class TileClassifier(torch.nn.Module):
    """Tile classifier for the tile optimised MaskRCNN model."""

    def __init__(self):
        super().__init__()
        self.fp16_enabled = False
        self.features = nn.Sequential(
            ConvModule(3, 64, 11, stride=4, padding=2, act_cfg=dict(type="ReLU")),
            nn.MaxPool2d(kernel_size=3, stride=2),
            ConvModule(64, 192, 5, padding=2, act_cfg=dict(type="ReLU")),
            nn.MaxPool2d(kernel_size=3, stride=2),
            ConvModule(192, 256, 3, padding=1, act_cfg=dict(type="ReLU")),
            nn.MaxPool2d(kernel_size=3, stride=2),
        )
        self.avgpool = torch.nn.AdaptiveAvgPool2d((6, 6))
        self.classifier = torch.nn.Sequential(
            torch.nn.Linear(256 * 6 * 6, 256),
            torch.nn.ReLU(inplace=True),
            torch.nn.Linear(256, 256),
            torch.nn.ReLU(inplace=True),
            torch.nn.Linear(256, 1),
        )

        self.loss_func = torch.nn.BCEWithLogitsLoss()
        self.sigmoid = torch.nn.Sigmoid()

    @auto_fp16()
    def forward(self, img: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            img (torch.Tensor): input image

        Returns:
            torch.Tensor: logits
        """
        x = self.features(img)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        y = self.classifier(x)
        return y

    @auto_fp16()
    def loss(self, pred: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Calculate BCE loss.

        Args:
            pred (torch.Tensor): logits
            target (torch.Tensor): binary target

        Returns:
            torch.Tensor: BCE loss
        """
        loss = self.loss_func(pred, target)
        return loss

    @auto_fp16()
    def simple_test(self, img: torch.Tensor) -> torch.Tensor:
        """Simple test.

        Args:
            img (torch.Tensor): input image

        Returns:
            torch.Tensor: objectness score
        """

        out = self.forward(img)
        with no_nncf_trace():
            return self.sigmoid(out)[0][0]


# pylint: disable=too-many-ancestors
@DETECTORS.register_module()
class CustomMaskRCNNTileOptimized(CustomMaskRCNN):
    """Custom MaskRCNN detector with tile classifier.

    Args:
        *args: args
        **kwargs: kwargs
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tile_classifier = TileClassifier()

    # pylint: disable=too-many-arguments
    def forward_train(
        self, img, img_metas, gt_bboxes, gt_labels, gt_bboxes_ignore=None, gt_masks=None, proposals=None, **kwargs
    ):
        """Forward pass during training.

        Joint training of tile classifier and MaskRCNN.

        Args:
            img (torch.Tensor): input image
            img_metas (list): image meta data
            gt_bboxes (list): ground truth bounding boxes
            gt_labels (list): ground truth labels
            gt_bboxes_ignore (list, optional): ground truth bounding boxes to be ignored. Defaults to None.
            gt_masks (list, optional): ground truth masks. Defaults to None.
            proposals (list, optional): proposals. Defaults to None.
            kwargs: kwargs
        """

        losses = dict()
        targets = [len(gt_bbox) > 0 for gt_bbox in gt_bboxes]
        pred = self.tile_classifier(img)
        target_labels = torch.tensor(targets, device=pred.device)
        loss_tile = self.tile_classifier.loss(pred, target_labels.unsqueeze(1).float())

        losses.update(dict(loss_tile=loss_tile))

        if not any(targets):
            return losses

        img = img[targets]
        img_metas = [item for keep, item in zip(targets, img_metas) if keep]
        gt_labels = [item for keep, item in zip(targets, gt_labels) if keep]
        gt_bboxes = [item for keep, item in zip(targets, gt_bboxes) if keep]
        gt_masks = [item for keep, item in zip(targets, gt_masks) if keep]
        if gt_bboxes_ignore:
            gt_bboxes_ignore = [item for keep, item in zip(targets, gt_bboxes_ignore) if keep]
        rcnn_loss = super().forward_train(
            img, img_metas, gt_bboxes, gt_labels, gt_bboxes_ignore, gt_masks, proposals, **kwargs
        )
        losses.update(rcnn_loss)
        return losses

    def simple_test(self, img, img_metas, proposals=None, rescale=False):
        """Simple test.

        Tile classifier is used to filter out images without any objects.
        If no objects are present, empty results are returned. Otherwise, MaskRCNN is used to detect objects.

        Args:
            img (torch.Tensor): input image
            img_metas (list): image meta data
            proposals (list, optional): proposals. Defaults to None.
            rescale (bool, optional): rescale. Defaults to False.

        Returns:
            tuple: MaskRCNN output
        """
        keep = self.tile_classifier.simple_test(img) > 0.45

        if not keep:
            tmp_results = []
            num_classes = 1
            bbox_results = []
            mask_results = []
            for _ in range(num_classes):
                bbox_results.append(np.empty((0, 5), dtype=np.float32))
                mask_results.append([])
            tmp_results.append((bbox_results, mask_results))
            return tmp_results

        assert self.with_bbox, "Bbox head must be implemented."
        x = self.extract_feat(img)

        if proposals is None:
            proposal_list = self.rpn_head.simple_test_rpn(x, img_metas)
        else:
            proposal_list = proposals

        return self.roi_head.simple_test(x, proposal_list, img_metas, rescale=rescale)


if is_mmdeploy_enabled():
    from mmdeploy.core import FUNCTION_REWRITER, mark
    from mmdeploy.utils import is_dynamic_shape

    # pylint: disable=ungrouped-imports
    from otx.algorithms.common.adapters.mmcv.hooks.recording_forward_hook import (
        ActivationMapHook,
        FeatureVectorHook,
    )

    @mark("tile_classifier", inputs=["image"], outputs=["tile_prob"])
    def tile_classifier__simple_test_impl(self, img):
        """Tile Classifier Simple Test Impl with added mmdeploy marking for model partitioning.

        Partition tile classifier by marking tile classifier in tracing.

        Args:
            self (object): object
            img: input image

        Returns:
            torch.Tensor: objectness score
        """
        return self.sigmoid(self.forward(img))[0][0]

    # pylint: disable=line-too-long, unused-argument
    @FUNCTION_REWRITER.register_rewriter(
        "otx.algorithms.detection.adapters.mmdet.models.detectors.custom_maskrcnn_tile_optimized.TileClassifier.simple_test"  # noqa: E501
    )
    def tile_classifier__simple_test(ctx, self, img):
        """Tile Classifier Simple Test Rewriter.

        Partition tile classifier by rewriting tile classifier simple test.

        Args:
            ctx (object): object context
            self (object): object
            img (torch.Tensor): input image

        Returns:
            torch.Tensor: objectness score
        """
        return tile_classifier__simple_test_impl(self, img)

    # pylint: disable=protected-access
    @mark(
        "custom_maskrcnn_forward",
        inputs=["input"],
        outputs=["dets", "labels", "masks", "tile_prob", "feats", "saliencies"],
    )
    def __forward_impl(ctx, self, img, img_metas, **kwargs):
        """Custom MaskRCNN Forward Impl with added mmdeploy marking for model partitioning.

        Args:
            ctx (object): object context
            self (object): object
            img (torch.Tensor): input image
            img_metas (dict): image meta data
            **kwargs: kwargs

        Returns:
            simple test: MaskRCNN output
        """
        assert isinstance(img, torch.Tensor)

        deploy_cfg = ctx.cfg
        is_dynamic_flag = is_dynamic_shape(deploy_cfg)
        # get origin input shape as tensor to support onnx dynamic shape
        img_shape = torch._shape_as_tensor(img)[2:]
        if not is_dynamic_flag:
            img_shape = [int(val) for val in img_shape]
        img_metas[0]["img_shape"] = img_shape
        return self.simple_test(img, img_metas, **kwargs)

    # pylint: disable=line-too-long
    @FUNCTION_REWRITER.register_rewriter(
        "otx.algorithms.detection.adapters.mmdet.models.detectors.custom_maskrcnn_tile_optimized.CustomMaskRCNNTileOptimized.forward"  # noqa: E501
    )
    def custom_maskrcnn__forward(ctx, self, img, img_metas=None, **kwargs):
        """Custom MaskRCNN Forward Rewriter.

        Args:
            ctx (object): object context
            self (object): object
            img (torch.Tensor): input image
            img_metas (dict, optional): image meta data. Defaults to None.
            **kwargs: kwargs

        Returns:
            MaskRCNN output
        """
        if img_metas is None:
            img_metas = [{}]
        else:
            assert len(img_metas) == 1, "do not support aug_test"
            img_metas = img_metas[0]

        if isinstance(img, list):
            img = img[0]

        return __forward_impl(ctx, self, img, img_metas=img_metas, **kwargs)

    # pylint: disable=line-too-long
    @FUNCTION_REWRITER.register_rewriter(
        "otx.algorithms.detection.adapters.mmdet.models.detectors.custom_maskrcnn_tile_optimized.CustomMaskRCNNTileOptimized.simple_test"  # noqa: E501
    )
    def custom_mask_rcnn__simple_test(ctx, self, img, img_metas, proposals=None):
        """Custom Mask RCNN Simple Test Rewriter for ONNX tracing.

        Tile classifier is added to ONNX tracing in order to partition the model.

        Args:
            ctx (object): object context
            self (object): object
            img (torch.Tensor): input image
            img_metas (list): image meta data
            proposals (list, optional): proposals. Defaults to None.

        Returns:
            tuple: MaskRCNN output with tile classifier output
        """
        assert self.with_bbox, "Bbox head must be implemented."
        tile_prob = self.tile_classifier.simple_test(img)

        x = self.extract_feat(img)
        if proposals is None:
            proposals, _ = self.rpn_head.simple_test_rpn(x, img_metas)
        out = self.roi_head.simple_test(x, proposals, img_metas, rescale=False)

        if ctx.cfg["dump_features"]:
            feature_vector = FeatureVectorHook.func(x)
            saliency_map = ActivationMapHook.func(x[-1])
            return (*out, tile_prob, feature_vector, saliency_map)

        return (*out, tile_prob)



"""A thin wrapper for storage cache using datumaro arrow format exporter."""
# copyright (c) 2023 intel corporation
# spdx-license-identifier: apache-2.0
#

import hashlib
import os
import shutil
import stat
from typing import List, Optional

from datumaro.components.dataset import Dataset as DatumDataset
from datumaro.components.progress_reporting import SimpleProgressReporter

from otx.core.file import OTX_CACHE

DATASET_CACHE = os.path.join(OTX_CACHE, "dataset")


def arrow_cache_helper(
    dataset: DatumDataset,
    scheme: str,
    num_workers: int = 0,
    cache_dir: str = DATASET_CACHE,
    force: bool = False,
) -> List[str]:
    """A helper for dumping Datumaro arrow format.

    Args:
        dataset: Datumaro dataset to export in apache arrow.
        scheme: Datumaro apache arrow image encoding scheme.
        num_workers: The number of workers to build arrow format.
        cache_dir: The directory to save.
        force: If true, rebuild arrow even if cache is hit.
    """

    def get_hash(dataset, scheme):
        source_path = dataset.data_path
        _hash = hashlib.sha256()
        _hash.update(f"{source_path}".encode("utf-8"))
        _hash.update(f"{len(dataset)}".encode("utf-8"))
        _hash.update(f"{scheme}".encode("utf-8"))
        if source_path:
            _hash.update(str(os.stat(source_path)[stat.ST_MTIME]).encode("utf-8"))
            if os.path.isdir(source_path):
                for root, dirs, files in os.walk(source_path):
                    for file in files:
                        file = os.path.join(root, file)
                        _hash.update(str(os.stat(file)[stat.ST_MTIME]).encode("utf-8"))
                    for _dir in dirs:
                        _dir = os.path.join(root, _dir)
                        _hash.update(str(os.stat(_dir)[stat.ST_MTIME]).encode("utf-8"))
        return _hash.hexdigest()

    def get_file_hash(file):
        _hash = hashlib.sha256()
        _hash.update(str(file).encode("utf-8"))
        _hash.update(str(os.stat(file)[stat.ST_MTIME]).encode("utf-8"))
        return _hash.hexdigest()

    cache_dir = os.path.join(cache_dir, get_hash(dataset, scheme))
    if os.path.exists(cache_dir) and force:
        shutil.rmtree(cache_dir)
    os.makedirs(cache_dir, exist_ok=True)

    cache_paths = []
    cache_hit = []
    for cache_path in os.listdir(cache_dir):
        if not cache_path.endswith(".arrow"):
            continue
        cache_hit.append(False)
        cache_path = os.path.join(cache_dir, cache_path)
        cache_paths.append(cache_path)
        hash_path = f"{cache_path}.hash"
        if os.path.exists(cache_path) and os.path.exists(hash_path):
            with open(hash_path, "r", encoding="utf-8") as f:
                if get_file_hash(cache_path) == f.read():
                    cache_hit[-1] = True

    if cache_hit and all(cache_hit):
        return cache_paths

    dataset.export(
        cache_dir,
        "arrow",
        save_media=True,
        image_ext=scheme,
        num_workers=num_workers,
        progress_reporter=SimpleProgressReporter(0, 10),
    )

    cache_paths = []
    for cache_path in os.listdir(cache_dir):
        if not cache_path.endswith(".arrow"):
            continue
        cache_path = os.path.join(cache_dir, cache_path)
        cache_paths.append(cache_path)
        hash_path = f"{cache_path}.hash"
        with open(hash_path, "w", encoding="utf-8") as f:
            f.write(get_file_hash(cache_path))

    return cache_paths


def init_arrow_cache(dataset: DatumDataset, scheme: Optional[str] = None, **kwargs) -> DatumDataset:
    """Init arrow format cache from Datumaro.

    Args:
        dataset: Datumaro dataset
        scheme: Datumaro apache arrow image encoding scheme.
        kwargs: kwargs passed to 'arrow_cache_helper'
    """
    if scheme is None or scheme == "NONE":
        return dataset
    cache_paths = arrow_cache_helper(dataset, scheme, **kwargs)
    dataset = DatumDataset.import_from(os.path.dirname(cache_paths[0]), "arrow")
    return dataset



"""This module tests classes related to InferenceParameters"""

# Copyright (C) 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import dataclasses

import pytest

from otx.api.entities.inference_parameters import InferenceParameters
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestInferenceParameters:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_inference_parameters_members(self):
        """
        <b>Description:</b>
        To test InferenceParameters dataclass members

        <b>Input data:</b>
        Initialized instance of InferenceParameters class

        <b>Expected results:</b>

        <b>Steps</b>
        1. Create InferenceParameters
        2. Check members
        """
        infer_params = InferenceParameters()

        assert dataclasses.is_dataclass(infer_params)
        assert len(dataclasses.fields(infer_params)) == 5
        assert dataclasses.fields(infer_params)[0].name == "is_evaluation"
        assert dataclasses.fields(infer_params)[1].name == "update_progress"
        assert dataclasses.fields(infer_params)[2].name == "explainer"
        assert dataclasses.fields(infer_params)[3].name == "process_saliency_maps"
        assert dataclasses.fields(infer_params)[4].name == "explain_predicted_classes"
        assert type(infer_params.is_evaluation) is bool
        assert type(infer_params.process_saliency_maps) is bool
        assert type(infer_params.explain_predicted_classes) is bool
        assert callable(infer_params.update_progress)
        assert type(infer_params.explainer) is str
        with pytest.raises(AttributeError):
            str(infer_params.WRONG)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_inference_parameters_update_member(self):
        """
        <b>Description:</b>
        To test InferenceParameters dataclass members update

        <b>Input data:</b>
        Initialized instance of InferenceParameters class

        <b>Expected results:</b>

        <b>Steps</b>
        1. Create InferenceParameters
        2. Check members update
        """
        infer_params = InferenceParameters(False)
        assert infer_params.is_evaluation is False
        assert (
            infer_params.update_progress(-2147483648) is infer_params.update_progress(0)
            and infer_params.update_progress(2147483648) is None
        )

        infer_params = InferenceParameters(True)
        assert infer_params.is_evaluation is True
        assert (
            infer_params.update_progress(-2147483648) is infer_params.update_progress(0)
            and infer_params.update_progress(2147483648) is None
        )



# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import datetime
from typing import List

import pytest

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
    NullAnnotationSceneEntity,
)
from otx.api.entities.color import Color
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.utils.time_utils import now
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestAnnotation:

    rectangle = Rectangle(x1=0.5, x2=1.0, y1=0.0, y2=0.5)
    labels: List[ScoredLabel] = []
    annotation = Annotation(shape=rectangle, labels=labels)

    car = LabelEntity(
        id=ID(123456789),
        name="car",
        domain=Domain.DETECTION,
        color=Color(red=16, green=15, blue=56, alpha=255),
        is_empty=True,
    )
    person = LabelEntity(
        id=ID(987654321),
        name="person",
        domain=Domain.DETECTION,
        color=Color(red=11, green=18, blue=38, alpha=200),
        is_empty=False,
    )
    car_label = ScoredLabel(car)
    person_label = ScoredLabel(person)
    labels2 = [car_label, person_label]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_default_property(self):
        """
        <b>Description:</b>
        Check that Annotation can correctly return default property value

        <b>Input data:</b>
        Annotation class

        <b>Expected results:</b>
        Test passes if the Annotation return correct values

        <b>Steps</b>
        1. Create Annotation instances
        2. Check default values
        """

        annotation = self.annotation

        assert type(annotation.id_) == ID
        assert annotation.id_ is not None
        assert str(annotation.shape) == "Rectangle(x=0.5, y=0.0, width=0.5, height=0.5)"
        assert annotation.get_labels() == []

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_setters(self):
        """
        <b>Description:</b>
        Check that Annotation can correctly return modified property value

        <b>Input data:</b>
        Annotation class

        <b>Expected results:</b>
        Test passes if the Annotation return correct values

        <b>Steps</b>
        1. Create Annotation instances
        2. Set another values
        3. Check changed values
        """

        annotation = self.annotation
        ellipse = Ellipse(x1=0.5, y1=0.1, x2=0.8, y2=0.3)
        annotation.shape = ellipse
        annotation.id_ = ID(123456789)

        assert annotation.id_ == ID(123456789)
        assert annotation.shape == ellipse

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_magic_methods(self):
        """
        <b>Description:</b>
        Check Annotation __repr__, __eq__ methods

        <b>Input data:</b>
        Initialized instance of Annotation

        <b>Expected results:</b>
        Test passes if Annotation magic methods returns correct values

        <b>Steps</b>
        1. Create Annotation instances
        2. Check returning value of magic methods
        """

        annotation = self.annotation
        other_annotation = self.annotation

        point1 = Point(0.3, 0.1)
        point2 = Point(0.8, 0.3)
        point3 = Point(0.6, 0.2)
        points = [point1, point2, point3]
        third_annotation = Annotation(shape=Polygon(points=points), labels=self.labels)

        assert repr(annotation) == "Annotation(shape=Ellipse(x1=0.5, y1=0.1, x2=0.8, y2=0.3), labels=[], id=123456789)"
        assert annotation == other_annotation
        assert annotation != third_annotation
        assert annotation != str

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_get_labels(self):
        """
        <b>Description:</b>
        Check Annotation get_labels method

        <b>Input data:</b>
        Initialized instance of Annotation

        <b>Expected results:</b>
        Test passes if Annotation get_labels method returns correct values

        <b>Steps</b>
        1. Create Annotation instances
        2. Check returning value of get_labels method
        3. Check returning value of get_labels method with include_empty=True
        """
        annotation = Annotation(shape=self.rectangle, labels=self.labels2)

        assert "[ScoredLabel(987654321, name=person, probability=0.0, domain=DETECTION," in str(annotation.get_labels())
        assert "color=Color(red=11, green=18, blue=38, alpha=200), hotkey=" in str(annotation.get_labels())
        assert ", label_source=LabelSource(user_id='', model_id=ID(), model_storage_id=ID()))]" in str(
            annotation.get_labels()
        )

        assert "[ScoredLabel(123456789, name=car" in str(annotation.get_labels(include_empty=True))
        assert ", probability=0.0, domain=DETECTION," in str(annotation.get_labels(include_empty=True))
        assert "color=Color(red=16, green=15," in str(annotation.get_labels(include_empty=True))
        assert "blue=56, alpha=255), hotkey=," in str(annotation.get_labels(include_empty=True))
        assert "label_source=LabelSource(user_id='', model_id=ID(), model_storage_id=ID()))," in str(
            annotation.get_labels(include_empty=True)
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_get_label_ids(self):
        """
        <b>Description:</b>
        Check Annotation get_label_ids method

        <b>Input data:</b>
        Initialized instance of Annotation

        <b>Expected results:</b>
        Test passes if Annotation get_label_ids method returns correct values

        <b>Steps</b>
        1. Create Annotation instances
        2. Check returning value of get_label_ids method
        3. Check returning value of get_label_ids method with include_empty=True
        """

        annotation = Annotation(shape=self.rectangle, labels=self.labels2)

        assert annotation.get_label_ids() == {ID(987654321)}
        assert annotation.get_label_ids(include_empty=True) == {
            ID(987654321),
            ID(123456789),
        }

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_append_label(self):
        """
        <b>Description:</b>
        Check Annotation append_label method

        <b>Input data:</b>
        Initialized instance of Annotation

        <b>Expected results:</b>
        Test passes if Annotation append_label method correct appending label

        <b>Steps</b>
        1. Create Annotation instances
        2. Append label
        3. Check labels
        """

        annotation = self.annotation

        annotation.append_label(label=self.car_label)
        assert annotation.get_labels() == []  # car_label is empty

        annotation.append_label(label=self.person_label)
        assert "name=person" in str(annotation.get_labels())

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_set_labels(self):
        """
        <b>Description:</b>
        Check Annotation set_labels method

        <b>Input data:</b>
        Initialized instance of Annotation

        <b>Expected results:</b>
        Test passes if Annotation set_labels method correct setting label

        <b>Steps</b>
        1. Create Annotation instances
        2. Set labels
        3. Check labels
        """

        annotation = self.annotation
        assert annotation.get_labels() != []

        annotation.set_labels(labels=[])
        assert annotation.get_labels() == []

        annotation.set_labels(labels=self.labels2)
        assert "name=person" in str(annotation.get_labels())
        assert "name=car" not in str(annotation.get_labels())  # car_label is empty


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestAnnotationSceneKind:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_scene_kind(self):
        """
        <b>Description:</b>
        Check that AnnotationSceneKind Enum lenght

        <b>Input data:</b>
        AnnotationSceneKind class

        <b>Expected results:</b>
        Test passes if the AnnotationSceneKind return correct values

        <b>Steps</b>
        1. Create AnnotationSceneKind instances
        2. Check Enum lenght
        """

        annotation_scene_kind = AnnotationSceneKind
        assert len(annotation_scene_kind) == 6

        assert str(annotation_scene_kind(0)) == "NONE"
        assert str(annotation_scene_kind(1)) == "ANNOTATION"
        assert str(annotation_scene_kind(2)) == "PREDICTION"
        assert str(annotation_scene_kind(3)) == "EVALUATION"
        assert str(annotation_scene_kind(4)) == "INTERMEDIATE"
        assert str(annotation_scene_kind(5)) == "TASK_PREDICTION"


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestAnnotationSceneEntity:

    creation_date = now()
    labels: List[ScoredLabel] = []
    rectangle = Rectangle(x1=0.5, x2=1.0, y1=0.0, y2=0.5)
    annotation = Annotation(shape=rectangle, labels=labels)

    point1 = Point(0.3, 0.1)
    point2 = Point(0.8, 0.3)
    point3 = Point(0.6, 0.2)
    points = [point1, point2, point3]
    polygon = Polygon(points=points)
    annotation2 = Annotation(shape=polygon, labels=labels)

    annotations = [annotation, annotation2]

    annotation_scene_entity = AnnotationSceneEntity(annotations=annotations, kind=AnnotationSceneKind.ANNOTATION)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_scene_entity_default_value(self):
        """
        <b>Description:</b>
        Check that AnnotationSceneEntity default values

        <b>Input data:</b>
        AnnotationSceneEntity class

        <b>Expected results:</b>
        Test passes if the AnnotationSceneEntity return correct values

        <b>Steps</b>
        1. Create AnnotationSceneEntity instances
        2. Check default values
        """

        annotation_scene_entity = self.annotation_scene_entity

        assert annotation_scene_entity.id_ == ID()
        assert annotation_scene_entity.kind == AnnotationSceneKind.ANNOTATION
        assert annotation_scene_entity.editor_name == ""
        assert type(annotation_scene_entity.creation_date) == datetime.datetime
        assert "Annotation(shape=Rectangle" in str(annotation_scene_entity.annotations)
        assert "Annotation(shape=Polygon" in str(annotation_scene_entity.annotations)
        assert annotation_scene_entity.shapes == [self.rectangle, self.polygon]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_scene_entity_setters(self):
        """
        <b>Description:</b>
        Check that AnnotationSceneEntity can correctly return modified property value

        <b>Input data:</b>
        Annotation class

        <b>Expected results:</b>
        Test passes if the AnnotationSceneEntity return correct values

        <b>Steps</b>
        1. Create AnnotationSceneEntity instances
        2. Set another values
        3. Check changed values
        """

        annotation_scene_entity = self.annotation_scene_entity

        creation_date = self.creation_date
        annotation_scene_entity.id_ = ID(123456789)
        annotation_scene_entity.kind = AnnotationSceneKind.PREDICTION
        annotation_scene_entity.editor_name = "editor"
        annotation_scene_entity.creation_date = creation_date
        annotation_scene_entity.annotations = self.annotation

        assert annotation_scene_entity.id_ == ID(123456789)
        assert annotation_scene_entity.kind == AnnotationSceneKind.PREDICTION
        assert annotation_scene_entity.editor_name == "editor"
        assert annotation_scene_entity.creation_date == creation_date
        assert annotation_scene_entity.annotations == self.annotation

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_scene_entity_magic_methods(self):
        """
        <b>Description:</b>
        Check Annotation __repr__ method

        <b>Input data:</b>
        Initialized instance of AnnotationSceneEntity

        <b>Expected results:</b>
        Test passes if AnnotationSceneEntity magic method returns correct values

        <b>Steps</b>
        1. Create AnnotationSceneEntity instances
        2. Check returning value of magic method
        """

        annotation_scene_entity = self.annotation_scene_entity

        annotation_scene_entity_repr = [
            f"{annotation_scene_entity.__class__.__name__}("
            f"annotations={annotation_scene_entity.annotations}, "
            f"kind={annotation_scene_entity.kind}, "
            f"editor={annotation_scene_entity.editor_name}, "
            f"creation_date={annotation_scene_entity.creation_date}, "
            f"id={annotation_scene_entity.id_})"
        ]

        for i in annotation_scene_entity_repr:
            assert i in repr(annotation_scene_entity)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_scene_entity_contains_any(self):
        """
        <b>Description:</b>
        Check Annotation contains_any method

        <b>Input data:</b>
        Initialized instance of AnnotationSceneEntity

        <b>Expected results:</b>
        Test passes if AnnotationSceneEntity contains_any method returns correct values

        <b>Steps</b>
        1. Create AnnotationSceneEntity instances
        2. Check returning value of contains_any method
        """

        annotation_scene_entity = self.annotation_scene_entity
        annotation_scene_entity.annotations = self.annotations

        car = LabelEntity(name="car", domain=Domain.DETECTION, is_empty=True)
        person = LabelEntity(name="person", domain=Domain.DETECTION)
        tree = LabelEntity(name="tree", domain=Domain.DETECTION)
        car_label = ScoredLabel(car)
        person_label = ScoredLabel(person)
        tree_label = ScoredLabel(tree)
        labels = [car_label]
        labels2 = [car_label, person_label]

        annotation = Annotation(shape=self.rectangle, labels=labels2)
        annotations = [annotation]
        annotation_scene_entity2 = AnnotationSceneEntity(annotations=annotations, kind=AnnotationSceneKind.ANNOTATION)

        assert annotation_scene_entity.contains_any(labels=labels) is False
        assert annotation_scene_entity2.contains_any(labels=labels2) is True
        assert annotation_scene_entity2.contains_any(labels=[tree_label]) is False

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_scene_entity_append_annotation(self):
        """
        <b>Description:</b>
        Check Annotation append_annotation method

        <b>Input data:</b>
        Initialized instance of AnnotationSceneEntity

        <b>Expected results:</b>
        Test passes if AnnotationSceneEntity append_annotation method returns correct values

        <b>Steps</b>
        1. Create AnnotationSceneEntity instances
        2. Check returning value of append_annotation method
        """

        annotation_scene_entity = self.annotation_scene_entity

        tree = LabelEntity(name="tree", domain=Domain.DETECTION)
        tree_label = ScoredLabel(tree)
        labels = [tree_label]
        annotation = Annotation(shape=self.rectangle, labels=labels)

        assert len(annotation_scene_entity.annotations) == 2

        annotation_scene_entity.append_annotation(annotation)
        assert len(annotation_scene_entity.annotations) == 3

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_scene_entity_append_annotations(self):
        """
        <b>Description:</b>
        Check Annotation append_annotations method

        <b>Input data:</b>
        Initialized instance of AnnotationSceneEntity

        <b>Expected results:</b>
        Test passes if AnnotationSceneEntity append_annotations method returns correct values

        <b>Steps</b>
        1. Create AnnotationSceneEntity instances
        2. Check returning value of append_annotations method
        """

        annotation_scene_entity = self.annotation_scene_entity

        annotation_scene_entity.append_annotations(self.annotations)
        assert len(annotation_scene_entity.annotations) == 6

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_scene_entity_get_labels(self):
        """
        <b>Description:</b>
        Check Annotation get_labels method

        <b>Input data:</b>
        Initialized instance of AnnotationSceneEntity

        <b>Expected results:</b>
        Test passes if AnnotationSceneEntity get_labels method returns correct values

        <b>Steps</b>
        1. Create AnnotationSceneEntity instances
        2. Check returning value of get_labels method
        """

        annotation_scene_entity = self.annotation_scene_entity

        assert len(annotation_scene_entity.get_labels()) == 1
        assert "name=tree" in str(annotation_scene_entity.get_labels())

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_annotation_scene_entity_get_label_ids(self):
        """
        <b>Description:</b>
        Check Annotation get_label_ids method

        <b>Input data:</b>
        Initialized instance of AnnotationSceneEntity

        <b>Expected results:</b>
        Test passes if AnnotationSceneEntity get_label_ids method returns correct values

        <b>Steps</b>
        1. Create AnnotationSceneEntity instances
        2. Check returning value of get_label_ids method
        """

        annotation_scene_entity = self.annotation_scene_entity

        assert annotation_scene_entity.get_label_ids() == {ID()}

        bus = LabelEntity(id=ID(123456789), name="bus", domain=Domain.DETECTION)
        bus_label = ScoredLabel(bus)
        labels = [bus_label]
        annotation = Annotation(shape=self.rectangle, labels=labels)
        annotation_scene_entity.append_annotation(annotation)

        assert annotation_scene_entity.get_label_ids() == {ID(), ID(123456789)}


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestNullAnnotationSceneEntity:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_null_annotation_scene_entity(self):
        """
        <b>Description:</b>
        Check that NullAnnotationSceneEntity

        <b>Input data:</b>
        NullAnnotationSceneEntity class

        <b>Expected results:</b>
        Test passes if the NullAnnotationSceneEntity return correct values

        <b>Steps</b>
        1. Create NullAnnotationSceneEntity instances
        2. Check default values
        """

        null_annotation = NullAnnotationSceneEntity()

        assert null_annotation.id_ == ID()
        assert null_annotation.kind == AnnotationSceneKind.NONE
        assert null_annotation.editor_name == ""
        assert type(null_annotation.creation_date) == datetime.datetime
        assert null_annotation.annotations == []
        assert repr(null_annotation) == "NullAnnotationSceneEntity()"
